import html
import html.entities
import itertools
import os
import sys
from math import e as _e, inf, log10, nan, pow
from numbers import Real
from typing import Any, Callable, Dict, Final, Iterable, List, Protocol, Sequence, TypeVar, Union, overload

__all__ = [
    "M_LOG10E",
    "T0",
    "c",
    "h",
    "k",
    "e",
    "CATALOG",
    "BUILD_TIME",
    "LINES",
    "FREQUENCY",
    "INTENSITY",
    "ID",
    "STRUCTURAL_FORMULA",
    "STOICHIOMETRIC_FORMULA",
    "MOLECULE",
    "MOLECULE_SYMBOL",
    "SPECIES_TAG",
    "NAME",
    "TRIVIAL_NAME",
    "ISOTOPOLOG",
    "STATE",
    "STATE_HTML",
    "INCHI_KEY",
    "DEGREES_OF_FREEDOM",
    "LOWER_STATE_ENERGY",
    "CONTRIBUTOR",
    "VERSION",
    "DATE_OF_ENTRY",
    "HUMAN_READABLE",
    "ghz_to_mhz",
    "ghz_to_nm",
    "ghz_to_rec_cm",
    "mhz_to_ghz",
    "mhz_to_nm",
    "mhz_to_rec_cm",
    "nm_to_ghz",
    "nm_to_mhz",
    "nm_to_rec_cm",
    "rec_cm_to_ghz",
    "rec_cm_to_mhz",
    "rec_cm_to_nm",
    "rec_cm_to_meV",
    "rec_cm_to_j",
    "meV_to_rec_cm",
    "j_to_rec_cm",
    "log10_sq_nm_mhz_to_sq_nm_mhz",
    "log10_sq_nm_mhz_to_log10_cm_per_molecule",
    "log10_sq_nm_mhz_to_cm_per_molecule",
    "sq_nm_mhz_to_log10_sq_nm_mhz",
    "log10_cm_per_molecule_to_log10_sq_nm_mhz",
    "cm_per_molecule_to_log10_sq_nm_mhz",
    "sort_unique",
    "merge_sorted",
    "search_sorted",
    "within",
    "chem_html",
    "best_name",
    "remove_html",
    "wrap_in_html",
    "ensure_prefix",
    "save_catalog_to_file",
    "ReleaseInfo",
    "latest_release",
    "update_with_pip",
    "tag",
    "p_tag",
    "a_tag",
    "LineType",
    "LinesType",
    "CatalogEntryType",
    "CatalogType",
    "CatalogJSONType",
    "OldCatalogJSONType",
]

M_LOG10E: Final[float] = log10(_e)

T0: Final[float] = 300.00  # [K], see https://spec.jpl.nasa.gov/ftp/pub/catalog/doc/catdoc.pdf
k: Final[float] = 1.380649000e-23  # [J/K],  see https://physics.nist.gov/cgi-bin/cuu/Value?k
h: Final[float] = 6.626070150e-34  # [J/Hz], see https://physics.nist.gov/cgi-bin/cuu/Value?h
e: Final[float] = 1.602176634e-19  # [C],    see https://physics.nist.gov/cgi-bin/cuu/Value?e
c: Final[float] = 299_792_458.000  # [m/s],  see https://physics.nist.gov/cgi-bin/cuu/Value?c

CATALOG: Final[str] = "catalog"
BUILD_TIME: Final[str] = "build_time"
LINES: Final[str] = "lines"
FREQUENCY: Final[str] = "frequency"
INTENSITY: Final[str] = "intensity"
ID: Final[str] = "id"
MOLECULE: Final[str] = "molecule"
STRUCTURAL_FORMULA: Final[str] = "structuralformula"
STOICHIOMETRIC_FORMULA: Final[str] = "stoichiometricformula"
MOLECULE_SYMBOL: Final[str] = "moleculesymbol"
SPECIES_TAG: Final[str] = "speciestag"
NAME: Final[str] = "name"
TRIVIAL_NAME: Final[str] = "trivialname"
ISOTOPOLOG: Final[str] = "isotopolog"
STATE: Final[str] = "state"
STATE_HTML: Final[str] = "state_html"
INCHI_KEY: Final[str] = "inchikey"
CONTRIBUTOR: Final[str] = "contributor"
VERSION: Final[str] = "version"
DATE_OF_ENTRY: Final[str] = "dateofentry"
DEGREES_OF_FREEDOM: Final[str] = "degreesoffreedom"
LOWER_STATE_ENERGY: Final[str] = "lowerstateenergy"

HUMAN_READABLE: Final[dict[str, str]] = {
    CATALOG: "Catalog",
    LINES: "Lines",
    FREQUENCY: "Frequency",
    INTENSITY: "Intensity",
    ID: "ID",
    MOLECULE: "Molecule",
    STRUCTURAL_FORMULA: "Structural formula",
    STOICHIOMETRIC_FORMULA: "Stoichiometric formula",
    MOLECULE_SYMBOL: "Molecule symbol",
    SPECIES_TAG: "Species tag",
    NAME: "Name",
    TRIVIAL_NAME: "Trivial name",
    ISOTOPOLOG: "Isotopolog",
    STATE: "State (TeX)",
    STATE_HTML: "State (HTML)",
    INCHI_KEY: "InChI key",
    CONTRIBUTOR: "Contributor",
    VERSION: "Version",
    DATE_OF_ENTRY: "Date of entry",
    DEGREES_OF_FREEDOM: "Degrees of freedom",
    LOWER_STATE_ENERGY: "Lower state energy",
}


class LineType:
    __slots__ = [FREQUENCY, INTENSITY, LOWER_STATE_ENERGY]

    def __init__(
        self,
        frequency: float = nan,
        intensity: float = nan,
        lowerstateenergy: float = nan,
    ) -> None:
        self.frequency: float = frequency
        self.intensity: float = intensity
        self.lowerstateenergy: float = lowerstateenergy


LinesType = List[LineType]


# noinspection PyShadowingBuiltins
class CatalogEntryType:
    __slots__ = [
        ID,
        MOLECULE,
        STRUCTURAL_FORMULA,
        STOICHIOMETRIC_FORMULA,
        MOLECULE_SYMBOL,
        SPECIES_TAG,
        NAME,
        TRIVIAL_NAME,
        ISOTOPOLOG,
        STATE,
        STATE_HTML,
        INCHI_KEY,
        CONTRIBUTOR,
        VERSION,
        DATE_OF_ENTRY,
        DEGREES_OF_FREEDOM,
        LINES,
    ]

    def __init__(
        self,
        id: int = 0,
        molecule: int = 0,
        structuralformula: str = "",
        stoichiometricformula: str = "",
        moleculesymbol: str = "",
        speciestag: int = 0,
        name: str = "",
        trivialname: str = "",
        isotopolog: str = "",
        state: str = "",
        state_html: str = "",
        inchikey: str = "",
        contributor: str = "",
        version: str = "",
        dateofentry: str = "",
        degreesoffreedom: int = -1,
        lines: LinesType = (),
    ) -> None:
        self.id: int = id
        self.molecule: int = molecule
        self.structuralformula: str = structuralformula
        self.stoichiometricformula: str = stoichiometricformula
        self.moleculesymbol: str = moleculesymbol
        self.speciestag: int = speciestag
        self.name: str = name
        self.trivialname: str = trivialname
        self.isotopolog: str = isotopolog
        self.state: str = state
        self.state_html: str = state_html
        self.inchikey: str = inchikey
        self.contributor: str = contributor
        self.version: str = version
        self.dateofentry: str = dateofentry
        self.degreesoffreedom: int = degreesoffreedom
        self.lines: LinesType = lines


CatalogType = Dict[int, CatalogEntryType]
CatalogJSONEntryType = Dict[str, Union[int, str, List[Dict[str, float]]]]
CatalogJSONType = Dict[str, CatalogJSONEntryType]
OldCatalogJSONType = List[CatalogJSONEntryType]


def within(x: float, limits: tuple[float, float] | tuple[tuple[float, float], ...]) -> bool:
    if len(limits) < 2:
        raise ValueError("Invalid limits")
    if all(isinstance(limit, Real) for limit in limits):
        return min(limits) <= x <= max(limits)
    elif all(isinstance(limit, tuple) for limit in limits):
        return any(min(limit) <= x <= max(limit) for limit in limits)
    else:
        raise TypeError("Invalid limits type")


_AnyType = TypeVar("_AnyType")


class SupportsLessAndEqual(Protocol[_AnyType]):
    def __eq__(self, other: _AnyType) -> bool:
        pass

    def __lt__(self, other: _AnyType) -> bool:
        pass

    def __le__(self, other: _AnyType) -> bool:
        pass


@overload
def sort_unique(
    items: Sequence[_AnyType], *, key: Callable[[_AnyType], SupportsLessAndEqual], reverse: bool = False
) -> list[_AnyType]:
    pass


@overload
def sort_unique(
    items: Sequence[SupportsLessAndEqual],
    *,
    key: Callable[[SupportsLessAndEqual], SupportsLessAndEqual] | None = None,
    reverse: bool = False,
) -> list[SupportsLessAndEqual]:
    pass


def sort_unique(
    items: Sequence[SupportsLessAndEqual] | Sequence[_AnyType],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
    reverse: bool = False,
) -> list[SupportsLessAndEqual]:
    sorted_items: list[SupportsLessAndEqual] = sorted(items, key=key, reverse=reverse)
    i: int = 0
    while i < len(sorted_items) - 1:
        while i < len(sorted_items) - 1 and sorted_items[i] == sorted_items[i + 1]:
            del sorted_items[i + 1]
        i += 1
    return sorted_items


@overload
def merge_sorted(
    items_1: Sequence[_AnyType], items_2: Sequence[_AnyType], *, key: Callable[[_AnyType], SupportsLessAndEqual]
) -> list[_AnyType]:
    pass


@overload
def merge_sorted(
    items_1: Sequence[SupportsLessAndEqual],
    items_2: Sequence[SupportsLessAndEqual],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
) -> list[SupportsLessAndEqual]:
    pass


def merge_sorted(
    items_1: Sequence[SupportsLessAndEqual] | Sequence[_AnyType],
    items_2: Sequence[SupportsLessAndEqual] | Sequence[_AnyType],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
) -> list[SupportsLessAndEqual]:
    sorted_items_1: list[SupportsLessAndEqual] = sort_unique(items_1, key=key, reverse=False)
    sorted_items_2: list[SupportsLessAndEqual] = sort_unique(items_2, key=key, reverse=False)
    merged_items: list[SupportsLessAndEqual] = []

    last_i_1: int
    last_i_2: int
    i_1: int = 0
    i_2: int = 0

    if key is None:

        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    while i_1 < len(sorted_items_1) and i_2 < len(sorted_items_2):
        last_i_1 = i_1
        while (
            i_1 < len(sorted_items_1)
            and i_2 < len(sorted_items_2)
            and key(sorted_items_1[i_1]) <= key(sorted_items_2[i_2])
        ):
            i_1 += 1
        if last_i_1 < i_1:
            merged_items.extend(sorted_items_1[last_i_1:i_1])

        last_i_2 = i_2
        while (
            i_1 < len(sorted_items_1)
            and i_2 < len(sorted_items_2)
            and key(sorted_items_2[i_2]) <= key(sorted_items_1[i_1])
        ):
            i_2 += 1
        if last_i_2 < i_2:
            merged_items.extend(sorted_items_2[last_i_2:i_2])

    while i_1 < len(sorted_items_1) and key(merged_items[-1]) == key(sorted_items_1[i_1]):
        i_1 += 1
    if i_1 < len(sorted_items_1) and i_2 >= len(sorted_items_2):
        merged_items.extend(sorted_items_1[i_1:])
    while i_2 < len(sorted_items_2) and key(merged_items[-1]) == key(sorted_items_2[i_2]):
        i_2 += 1
    if i_2 < len(sorted_items_2) and i_1 >= len(sorted_items_1):
        merged_items.extend(sorted_items_2[i_2:])
    return merged_items


@overload
def search_sorted(
    threshold: SupportsLessAndEqual,
    items: Sequence[_AnyType],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual],
    maybe_equal: bool = False,
) -> int:
    pass


@overload
def search_sorted(
    threshold: SupportsLessAndEqual,
    items: Sequence[SupportsLessAndEqual],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
    maybe_equal: bool = False,
) -> int:
    pass


def search_sorted(
    threshold: SupportsLessAndEqual,
    items: Sequence[_AnyType] | Sequence[SupportsLessAndEqual],
    *,
    key: Callable[[_AnyType], SupportsLessAndEqual] | None = None,
    maybe_equal: bool = False,
) -> int:
    from operator import lt, le

    if not items:
        raise ValueError("Empty sequence provided")
    if key is None:

        def key(value: _AnyType) -> SupportsLessAndEqual:
            return value

    less: Callable[[SupportsLessAndEqual, SupportsLessAndEqual], bool] = le if maybe_equal else lt
    if not less(key(items[0]), threshold):
        return -1
    if less(key(items[-1]), threshold):
        return len(items)
    i: int = 0
    j: int = len(items) - 1
    n: int
    while j - i > 1:
        n = (i + j) // 2
        if less(key(items[n]), threshold):
            i = n
        else:
            j = n
    if i != j and not less(key(items[j]), threshold):
        return i
    return j


def mhz_to_ghz(frequency_mhz: float) -> float:
    return frequency_mhz * 1e-3


def mhz_to_rec_cm(frequency_mhz: float) -> float:
    return frequency_mhz * 1e4 / c


def mhz_to_nm(frequency_mhz: float) -> float:
    return c / frequency_mhz * 1e3


def ghz_to_mhz(frequency_ghz: float) -> float:
    return frequency_ghz * 1e3


def ghz_to_rec_cm(frequency_ghz: float) -> float:
    return frequency_ghz * 1e7 / c


def ghz_to_nm(frequency_ghz: float) -> float:
    return c / frequency_ghz


def rec_cm_to_mhz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-4 * c


def rec_cm_to_ghz(frequency_rec_cm: float) -> float:
    return frequency_rec_cm * 1e-7 * c


def rec_cm_to_nm(frequency_rec_cm: float) -> float:
    return 1e7 / frequency_rec_cm


def rec_cm_to_meV(energy_rec_cm: float) -> float:
    return 1e5 * h * c / e * energy_rec_cm


def rec_cm_to_j(energy_rec_cm: float) -> float:
    return 1e2 * h * c * energy_rec_cm


def nm_to_mhz(frequency_nm: float) -> float:
    return c / frequency_nm * 1e-3


def nm_to_ghz(frequency_nm: float) -> float:
    return c / frequency_nm


def nm_to_rec_cm(frequency_nm: float) -> float:
    return 1e7 / frequency_nm


def meV_to_rec_cm(energy_mev: float) -> float:
    return 1e-5 * e / h / c * energy_mev


def j_to_rec_cm(energy_j: float) -> float:
    return 1e-2 / h / c * energy_j


def log10_sq_nm_mhz_to_sq_nm_mhz(intensity_log10_sq_nm_mhz: float) -> float:
    return pow(10.0, intensity_log10_sq_nm_mhz)


def log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return -10.0 + intensity_log10_sq_nm_mhz - log10(c)


def log10_sq_nm_mhz_to_cm_per_molecule(intensity_log10_sq_nm_mhz: float) -> float:
    return pow(10.0, log10_sq_nm_mhz_to_log10_cm_per_molecule(intensity_log10_sq_nm_mhz))


def sq_nm_mhz_to_log10_sq_nm_mhz(intensity_sq_nm_mhz: float) -> float:
    if intensity_sq_nm_mhz == 0.0:
        return -inf
    if intensity_sq_nm_mhz < 0.0:
        return nan
    return log10(intensity_sq_nm_mhz)


def log10_cm_per_molecule_to_log10_sq_nm_mhz(intensity_log10_cm_per_molecule: float) -> float:
    return intensity_log10_cm_per_molecule + 10.0 + log10(c)


def cm_per_molecule_to_log10_sq_nm_mhz(intensity_cm_per_molecule: float) -> float:
    if intensity_cm_per_molecule == 0.0:
        return -inf
    if intensity_cm_per_molecule < 0.0:
        return nan
    return log10_cm_per_molecule_to_log10_sq_nm_mhz(log10(intensity_cm_per_molecule))


def tex_to_html_entity(s: str) -> str:
    r"""
    Change LaTeX entities syntax to HTML one.
    Get ‘\alpha’ to be ‘&alpha;’ and so on.
    Unknown LaTeX entities do not get replaced.

    :param s: A line to convert
    :return: a line with all LaTeX entities renamed
    """
    word_start: int = -1
    word_started: bool = False
    backslash_found: bool = False
    _i: int = 0
    fixes: dict[str, str] = {
        "neq": "#8800",
    }
    while _i < len(s):
        _c: str = s[_i]
        if word_started and not _c.isalpha():
            word_started = False
            if s[word_start:_i] + ";" in html.entities.entitydefs:
                s = s[: word_start - 1] + "&" + s[word_start:_i] + ";" + s[_i:]
                _i += 2
            elif s[word_start:_i] in fixes:
                s = s[: word_start - 1] + "&" + fixes[s[word_start:_i]] + ";" + s[_i:]
                _i += 2
        if backslash_found and _c.isalpha() and not word_started:
            word_start = _i
            word_started = True
        backslash_found = _c == "\\"
        _i += 1
    if word_started:
        if s[word_start:_i] + ";" in html.entities.entitydefs:
            s = s[: word_start - 1] + "&" + s[word_start:_i] + ";" + s[_i:]
            _i += 2
        elif s[word_start:_i] in fixes:
            s = s[: word_start - 1] + "&" + fixes[s[word_start:_i]] + ";" + s[_i:]
            _i += 2
    return s


def chem_html(formula: str) -> str:
    """converts plain text chemical formula into html markup"""
    if "<" in formula or ">" in formula:
        # we can not tell whether it's a tag or a mathematical sign
        return formula

    def sub_tag(s: str) -> str:
        return "<sub>" + s + "</sub>"

    def sup_tag(s: str) -> str:
        return "<sup>" + s + "</sup>"

    def i_tag(s: str) -> str:
        return "<i>" + s + "</i>"

    def subscript(s: str) -> str:
        number_start: int = -1
        number_started: bool = False
        cap_alpha_started: bool = False
        low_alpha_started: bool = False
        _i: int = 0
        while _i < len(s):
            _c: str = s[_i]
            if number_started and not _c.isdigit():
                number_started = False
                s = s[:number_start] + sub_tag(s[number_start:_i]) + s[_i:]
                _i += 1
            if (cap_alpha_started or low_alpha_started) and _c.isdigit() and not number_started:
                number_start = _i
                number_started = True
            if low_alpha_started:
                cap_alpha_started = False
                low_alpha_started = False
            if cap_alpha_started and _c.islower() or _c == ")":
                low_alpha_started = True
            cap_alpha_started = _c.isupper()
            _i += 1
        if number_started:
            s = s[:number_start] + sub_tag(s[number_start:])
        return s

    def prefix(s: str) -> str:
        no_digits: bool = False
        _i: int = len(s)
        while not no_digits:
            _i = s.rfind("-", 0, _i)
            if _i == -1:
                break
            if s[:_i].isalpha() and s[:_i].isupper():
                break
            no_digits = True
            _c: str
            unescaped_prefix: str = html.unescape(s[:_i])
            for _c in unescaped_prefix:
                if _c.isdigit() or _c == "<":
                    no_digits = False
                    break
            if no_digits and (unescaped_prefix[0].islower() or unescaped_prefix[0] == "("):
                return i_tag(s[:_i]) + s[_i:]
        return s

    def charge(s: str) -> str:
        if s[-1] in "+-":
            return s[:-1] + sup_tag(s[-1])
        return s

    def v(s: str) -> str:
        if "=" not in s:
            return s[0] + " = " + s[1:]
        ss: list[str] = list(map(str.strip, s.split("=")))
        for _i in range(len(ss)):
            if ss[_i].startswith("v"):
                ss[_i] = ss[_i][0] + sub_tag(ss[_i][1:])
        return " = ".join(ss)

    html_formula: str = html.escape(formula)
    html_formula_pieces: list[str] = list(map(str.strip, html_formula.split(",")))
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith("v"):
            html_formula_pieces = html_formula_pieces[:i] + [", ".join(html_formula_pieces[i:])]
            break
    for i in range(len(html_formula_pieces)):
        if html_formula_pieces[i].startswith("v"):
            html_formula_pieces[i] = v(html_formula_pieces[i])
            break
        for function in (subscript, prefix, charge):
            html_formula_pieces[i] = function(html_formula_pieces[i])
    html_formula = ", ".join(html_formula_pieces)
    return html_formula


def is_good_html(text: str) -> bool:
    """Basic check that all tags are sound"""
    _1, _2, _3 = text.count("<"), text.count(">"), 2 * text.count("</")
    return _1 == _2 and _1 == _3


def best_name(entry: CatalogEntryType, allow_html: bool = True) -> str:
    species_tag: int = entry.speciestag
    last: str = best_name.__dict__.get("last", dict()).get(species_tag, dict()).get(allow_html, "")
    if last:
        return last

    def _best_name() -> str:
        if isotopolog := entry.isotopolog:
            if allow_html:
                if is_good_html(str(molecule_symbol := entry.moleculesymbol)) and (
                    entry.structuralformula == isotopolog or entry.stoichiometricformula == isotopolog
                ):
                    if state_html := entry.state_html:
                        # span tags are needed when the molecule symbol is malformed
                        return f"<span>{molecule_symbol}</span>, {chem_html(tex_to_html_entity(str(state_html)))}"
                    return str(molecule_symbol)
                else:
                    if state_html := entry.state_html:
                        return f"{chem_html(str(isotopolog))}, {chem_html(tex_to_html_entity(str(state_html)))}"
                    return chem_html(str(isotopolog))
            else:
                if state_html := entry.state_html:
                    return f"{isotopolog}, {remove_html(tex_to_html_entity(state_html))}"
                if state := entry.state:
                    return f"{isotopolog}, {remove_html(tex_to_html_entity(state.strip('$')))}"
                return isotopolog

        for key in (NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA):
            if candidate := getattr(entry, key, ""):
                return chem_html(str(candidate)) if allow_html else str(candidate)
        if trivial_name := entry.trivialname:
            return str(trivial_name)
        if species_tag:
            return str(species_tag)
        return "no name"

    res: str = _best_name()
    if not species_tag:
        return res
    if "last" not in best_name.__dict__:
        best_name.__dict__["last"] = dict()
    if species_tag not in best_name.__dict__["last"]:
        best_name.__dict__["last"][species_tag] = dict()
    best_name.__dict__["last"][species_tag][allow_html] = res
    return res


def remove_html(line: str) -> str:
    """removes HTML tags and decodes HTML entities"""
    if not is_good_html(line):
        return html.unescape(line)

    new_line: str = line
    tag_start: int = new_line.find("<")
    tag_end: int = new_line.find(">", tag_start)
    while tag_start != -1 and tag_end != -1:
        new_line = new_line[:tag_start] + new_line[tag_end + 1 :]
        tag_start = new_line.find("<")
        tag_end = new_line.find(">", tag_start)
    return html.unescape(new_line).lstrip()


def wrap_in_html(text: str, line_end: str = os.linesep) -> str:
    """Make a full HTML document out of a piece of the markup"""
    new_text: list[str] = [
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">',
        '<html lang="en" xml:lang="en">',
        "<head>",
        '<meta http-equiv="content-type" content="text/html; charset=utf-8">',
        "</head>",
        "<body>",
        text,
        "</body>",
        "</html>",
    ]

    return line_end.join(new_text)


def ensure_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text
    else:
        return prefix + text


def save_catalog_to_file(
    filename: str | os.PathLike[str],
    catalog: CatalogType,
    frequency_limits: tuple[float, float],
) -> bool:
    from .catalog import Catalog

    if not catalog:
        return False
    Catalog.from_data(catalog_data=catalog, frequency_limits=frequency_limits).save(filename=filename)
    return True


class ReleaseInfo:
    def __init__(self, version: str = "", pub_date: str = "") -> None:
        self.version: str = version
        self.pub_date: str = pub_date

    def __bool__(self) -> bool:
        return bool(self.version) and bool(self.pub_date)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (str, ReleaseInfo)):
            raise TypeError("The argument must be a string or ReleaseInfo")
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        i: str
        j: str
        for i, j in itertools.zip_longest(
            self.version.replace("-", ".").split("."), other.version.replace("-", ".").split("."), fillvalue=""
        ):
            if i == j:
                continue
            if i.isdigit() and j.isdigit():
                return int(i) < int(j)
            else:
                i_digits: str = "".join(itertools.takewhile(str.isdigit, i))
                j_digits: str = "".join(itertools.takewhile(str.isdigit, j))
                if i_digits != j_digits:
                    if i_digits and j_digits:
                        return int(i_digits) < int(j_digits)
                    else:
                        return i_digits < j_digits
                return i < j
        return False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, (str, ReleaseInfo)):
            raise TypeError("The argument must be a string or ReleaseInfo")
        if isinstance(other, str):
            other = ReleaseInfo(version=other)
        return self.version == other.version


def latest_release() -> ReleaseInfo:
    import urllib.request
    import xml.dom.minidom as dom
    from http.client import HTTPResponse
    from xml.dom.minicompat import NodeList
    from urllib.error import URLError

    from . import __original_name__

    try:
        r: HTTPResponse = urllib.request.urlopen(
            f"https://pypi.org/rss/project/{__original_name__}/releases.xml", timeout=1
        )
    except URLError:
        return ReleaseInfo()
    if r.getcode() != 200 or not r.readable():
        return ReleaseInfo()
    rss: dom.Node | None = dom.parseString(r.read().decode(encoding="ascii")).documentElement
    if not isinstance(rss, dom.Element) or rss.tagName != "rss":
        return ReleaseInfo()
    channels: NodeList = rss.getElementsByTagName("channel")
    if not channels or channels[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    channel: dom.Element = channels[0]
    items: NodeList = channel.getElementsByTagName("item")
    if not items or items[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    item: dom.Element = items[0]
    titles: NodeList = item.getElementsByTagName("title")
    if not titles or titles[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    title: dom.Element = titles[0]
    pub_dates: NodeList = item.getElementsByTagName("pubDate")
    if not pub_dates or pub_dates[0].nodeType != dom.Node.ELEMENT_NODE:
        return ReleaseInfo()
    pub_date: dom.Element = pub_dates[0]
    title_value: dom.Node = title.firstChild
    pub_date_value: dom.Node = pub_date.firstChild
    if not isinstance(title_value, dom.Text) or not isinstance(pub_date_value, dom.Text):
        return ReleaseInfo()

    return ReleaseInfo(title_value.data, pub_date_value.data)


def update_with_pip() -> None:
    import subprocess
    import sys

    from . import __original_name__

    subprocess.Popen(
        args=[
            sys.executable,
            "-c",
            f"""import sys, subprocess, time; time.sleep(2);\
        subprocess.run(args=[sys.executable, '-m', 'pip', 'install', '-U', {__original_name__!r}]);\
        subprocess.Popen(args=[sys.executable, '-m', {__original_name__!r}])""",
        ]
    )
    sys.exit(0)


def tag(name: str, text: str = "", **attrs: str) -> str:
    parts: list[str] = ["<", " ".join((name, *itertools.starmap(lambda a, v: f"{a}={str(v)!r}", attrs.items())))]
    if text:
        parts.extend([">", text, "</", name, ">"])
    else:
        parts.append("/>")
    return "".join(parts)


def p_tag(text: str) -> str:
    return tag("p", text)


def a_tag(text: str, url: str) -> str:
    return tag("a", text, href=url)


if sys.version_info < (3, 10):
    # noinspection PyUnresolvedReferences
    import builtins

    # noinspection PyShadowingBuiltins, PyUnusedLocal
    def zip(*iterables: Iterable[Any], strict: bool = False) -> builtins.zip:
        """Intentionally override `builtins.zip` to ignore `strict` parameter in Python < 3.10"""
        return builtins.zip(*iterables)

    __all__.append("zip")



import enum
from pathlib import Path
from typing import Iterable, cast

from qtpy.QtCore import QDateTime, QLocale, QModelIndex, QTimeZone, QUrl, Qt, Signal, Slot
from qtpy.QtGui import QContextMenuEvent, QDesktopServices
from qtpy.QtWidgets import (
    QAbstractItemView,
    QAbstractScrollArea,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QMenu,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from . import qta_icon
from .selectable_label import SelectableLabel
from .settings import Settings
from .titled_list_widget import TitledListWidget
from .update_dialog import UpdateDialog
from ..catalog import Catalog, CatalogSourceInfo

__all__ = ["CatalogInfo"]


class SourcesList(QTableWidget):
    catalogUpdated: Signal = Signal(name="catalogUpdated")

    class Columns(enum.IntEnum):
        FileLocation = 0
        BuildTime = 1

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.settings: Settings = settings

        self.setAlternatingRowColors(True)
        self.setColumnCount(len(SourcesList.Columns.__members__))
        self.setCornerButtonEnabled(False)
        self.setDragDropOverwriteMode(False)
        self.setDropIndicatorShown(False)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setHorizontalHeaderLabels([self.tr("Filename"), self.tr("Build Time")])
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSortingEnabled(True)
        self.horizontalHeader().setHighlightSections(False)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setHighlightSections(False)

        self._context_menu: QMenu = QMenu(self)
        self._context_menu.setDefaultAction(
            self._context_menu.addAction(
                qta_icon("mdi6.target"), self.tr("Open File &Location"), self._on_open_file_location_triggered
            )
        )
        self._context_menu.addSeparator()
        self._context_menu.addAction(
            qta_icon("mdi6.update"), self.tr("&Update Catalog"), self._on_update_catalog_triggered
        )

        self.cellDoubleClicked.connect(self._on_item_double_clicked)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        if not self.selectedIndexes():
            return super().contextMenuEvent(event)
        self._context_menu.exec(event.globalPos())
        return super().contextMenuEvent(event)

    def _selected_row(self) -> int | None:
        if not self.selectedIndexes():
            return None
        return cast(QModelIndex, self.selectedIndexes()[0]).row()

    def _file_location(self, row: int) -> Path | None:
        item: QTableWidgetItem | None = self.item(row, SourcesList.Columns.FileLocation)
        if item is None:
            return None
        return item.data(Qt.ItemDataRole.UserRole)

    def _locate_file(self, row: int) -> bool:
        file_location: Path | None = self._file_location(row)
        if file_location is None:
            return False
        return QDesktopServices.openUrl(QUrl.fromLocalFile(str(file_location.parent)))

    @Slot(int, int)
    def _on_item_double_clicked(self, row: int, _column: int) -> None:
        self._locate_file(row)

    @Slot()
    def _on_open_file_location_triggered(self) -> None:
        row: int | None = self._selected_row()
        if row is None:
            return
        self._locate_file(row)

    @Slot()
    def _on_update_catalog_triggered(self) -> None:
        row: int | None = self._selected_row()
        if row is None:
            return
        file_location: Path | None = self._file_location(row)
        if file_location is None:
            return
        ud: UpdateDialog = UpdateDialog(
            settings=self.settings,
            existing_catalog_location=file_location,
            parent=self.parent(),
        )
        if ud.exec():
            self.catalogUpdated.emit()

    def extend(self, info: Iterable[CatalogSourceInfo]) -> None:
        row: int
        info_item: CatalogSourceInfo
        item: QTableWidgetItem
        for row, info_item in enumerate(info, start=self.rowCount()):
            self.setRowCount(row + 1)
            item = QTableWidgetItem(str(info_item.filename))
            item.setToolTip(str(info_item.filename))
            item.setData(Qt.ItemDataRole.UserRole, info_item.filename)
            self.setItem(row, SourcesList.Columns.FileLocation, item)
            if info_item.build_datetime is not None:
                qt_datetime: QDateTime = QDateTime(info_item.build_datetime)
                qt_datetime.setTimeZone(QTimeZone(round(info_item.build_datetime.utcoffset().total_seconds())))
                item = QTableWidgetItem(qt_datetime.toLocalTime().toString())
                self.setItem(row, SourcesList.Columns.BuildTime, item)
        self.setColumnHidden(
            SourcesList.Columns.BuildTime,
            all(self.item(row, SourcesList.Columns.BuildTime) is None for row in range(self.rowCount())),
        )
        self.resizeColumnsToContents()


class CatalogInfo(QDialog):
    """A simple dialog that displays the information about the loaded catalog(s)"""

    catalogUpdated: Signal = Signal(name="catalogUpdated")

    def __init__(self, settings: Settings, catalog: Catalog, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr("Catalog Info"))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)

        sources_list: SourcesList = SourcesList(settings=settings, parent=self)
        layout.addWidget(sources_list)
        sources_list.extend(catalog.sources_info)
        sources_list.catalogUpdated.connect(self._on_catalog_updated)

        frequency_limits_list: TitledListWidget = TitledListWidget(self)
        frequency_limits_list.setTitle(self.tr("Frequency limits:"))
        layout.addWidget(frequency_limits_list)
        locale: QLocale = self.locale()
        frequency_limits_list.addItems(
            [
                self.tr("{min_frequency} to {max_frequency} MHz").format(
                    min_frequency=locale.toString(min(frequency_limit)),
                    max_frequency=locale.toString(max(frequency_limit)),
                )
                for frequency_limit in catalog.frequency_limits
            ]
        )

        stat_layout: QFormLayout = QFormLayout()
        stat_layout.addRow(self.tr("Total number of substances:"), SelectableLabel(str(catalog.entries_count)))
        layout.addLayout(stat_layout, 0)

        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        self.adjustSize()

    @Slot()
    def _on_catalog_updated(self) -> None:
        self.reject()
        self.catalogUpdated.emit()



#!/usr/bin/env python3


def test_async_download():
    from pycatsearch import async_download

    async_download()


if __name__ == "__main__":
    import sys
    from os import path

    sys.path = list(set(sys.path) | {path.abspath(path.join(__file__, path.pardir, path.pardir))})

    test_async_download()



#!/usr/bin/env python3


def test_search():
    from tempfile import NamedTemporaryFile

    from pycatsearch.catalog import Catalog

    with NamedTemporaryFile("wb", suffix=".json") as f:
        f.write(
            b"""\
{
    "catalog": [
        {
            "id": 31,
            "molecule": 29,
            "structuralformula": "NH3",
            "stoichiometricformula": "H3N",
            "moleculesymbol": "NH<sub>3</sub>",
            "speciestag": 17004,
            "name": "NH3-v2",
            "trivialname": "Ammonia",
            "isotopolog": "NH3",
            "state": "$v2=1$",
            "state_html": "v2=1",
            "inchikey": "QGZKDVFQNNGYKY-UHFFFAOYSA-N",
            "contributor": "Shanshan Yu",
            "version": "5",
            "dateofentry": "2010-09-01",
            "degreesoffreedom": 3,
            "lines": [
                {
                    "frequency": 140141.8067,
                    "intensity": -5.0383,
                    "lowerstateenergy": 983.3773
                },
                {
                    "frequency": 140142.0001,
                    "intensity": -5.0383,
                    "lowerstateenergy": 983.3773
                }
            ]
        }
    ],
    "frequency": [
        110000.0,
        170000.0
    ],
    "build_time": "2023-05-05T10:07:28.690146+00:00"
}
"""
        )
        f.flush()
        c = Catalog(f.name)
    assert c, c.sources

    assert len(c.filter(min_frequency=140141, max_frequency=140142)[17004]["lines"]) == 1
    assert not c.filter(any_name_or_formula="oxygen")
    assert len(c.filter_by_species_tags(species_tags=[17004])[17004]["lines"]) == 2


if __name__ == "__main__":
    import sys
    from os import path

    sys.path = list(set(sys.path) | {path.abspath(path.join(__file__, path.pardir, path.pardir))})

    test_search()



from math import inf
from typing import Callable

from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import QAbstractSpinBox, QDoubleSpinBox, QFormLayout, QSizePolicy, QTabWidget, QWidget

from .settings import Settings

__all__ = ["FrequencyBox"]


class FrequencySpinBox(QDoubleSpinBox):
    """`QDoubleSpinBox` with altered defaults"""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignTrailing | Qt.AlignmentFlag.AlignVCenter)
        self.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setAccelerated(True)
        self.setDecimals(4)
        self.setMaximum(9999999.9999)
        self.setSuffix(self.tr(" MHz"))
        self.setCorrectionMode(QAbstractSpinBox.CorrectionMode.CorrectToNearestValue)
        self.setKeyboardTracking(False)  # not to emit signals while typing
        self.setSizePolicy(QSizePolicy.Policy.Expanding, self.sizePolicy().verticalPolicy())


class FrequencyBox(QTabWidget):
    frequencyLimitsChanged: Signal = Signal(name="frequencyLimitsChanged")

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        from . import qta_icon  # import locally to avoid a circular import

        super().__init__(parent)

        self._settings: Settings = settings

        self._frequency_from: float = -inf  # [MHz]
        self._frequency_to: float = inf  # [MHz]
        self._frequency_center: float = 0.0  # [MHz]
        self._frequency_deviation: float = inf  # [MHz]

        self._page_by_range: QWidget = QWidget()
        self._layout_by_range: QFormLayout = QFormLayout(self._page_by_range)
        self._spin_frequency_from: FrequencySpinBox = FrequencySpinBox(self._page_by_range)
        self._spin_frequency_to: FrequencySpinBox = FrequencySpinBox(self._page_by_range)
        self._page_by_center: QWidget = QWidget()
        self._layout_by_center: QFormLayout = QFormLayout(self._page_by_center)
        self._spin_frequency_center: FrequencySpinBox = FrequencySpinBox(self._page_by_center)
        self._spin_frequency_deviation: FrequencySpinBox = FrequencySpinBox(self._page_by_center)

        self._spin_frequency_from.setValue(118747.341)
        self._layout_by_range.addRow(self.tr("From:"), self._spin_frequency_from)
        self._spin_frequency_to.setValue(118753.341)
        self._layout_by_range.addRow(self.tr("To:"), self._spin_frequency_to)
        self.addTab(self._page_by_range, qta_icon("mdi6.arrow-expand-horizontal"), self.tr("Range"))

        self._spin_frequency_center.setValue(118750.341)
        self._layout_by_center.addRow(self.tr("Center:"), self._spin_frequency_center)
        self._spin_frequency_deviation.setMaximum(99.9999)
        self._spin_frequency_deviation.setSingleStep(0.1)
        self._spin_frequency_deviation.setValue(0.4)
        self._layout_by_center.addRow(self.tr("Deviation:"), self._spin_frequency_deviation)
        self.addTab(self._page_by_center, qta_icon("mdi6.format-horizontal-align-center"), self.tr("Center"))

        self.load_settings()

        self._spin_frequency_from.valueChanged.connect(self._on_spin_frequency_from_edited)
        self._spin_frequency_to.valueChanged.connect(self._on_spin_frequency_to_edited)
        self._spin_frequency_center.valueChanged.connect(self._on_spin_frequency_center_edited)
        self._spin_frequency_deviation.valueChanged.connect(self._on_spin_frequency_deviation_edited)

    def load_settings(self) -> None:
        with self._settings.section("search"), self._settings.section("frequency"):
            self._frequency_from = self._settings.value("from", self._spin_frequency_from.value(), float)
            self._frequency_to = self._settings.value("to", self._spin_frequency_to.value(), float)
            self._frequency_center = self._settings.value("center", self._spin_frequency_center.value(), float)
            self._frequency_deviation = self._settings.value("deviation", self._spin_frequency_deviation.value(), float)

    def save_settings(self) -> None:
        with self._settings.section("search"), self._settings.section("frequency"):
            self._settings.setValue("from", self._frequency_from)
            self._settings.setValue("to", self._frequency_to)
            self._settings.setValue("center", self._frequency_center)
            self._settings.setValue("deviation", self._frequency_deviation)

    @property
    def min_frequency(self) -> float:
        if self.currentWidget() is self._page_by_range:
            return self._frequency_from
        else:
            return self._frequency_center - self._frequency_deviation

    @property
    def max_frequency(self) -> float:
        if self.currentWidget() is self._page_by_range:
            return self._frequency_to
        else:
            return self._frequency_center + self._frequency_deviation

    def set_frequency_limits(self, min_value: float, max_value: float) -> None:
        frequency_spins: list[QDoubleSpinBox] = [
            self._spin_frequency_from,
            self._spin_frequency_to,
            self._spin_frequency_center,
        ]
        min_value = self._settings.from_mhz(min_value)
        max_value = self._settings.from_mhz(max_value)
        spin: QDoubleSpinBox
        for spin in frequency_spins:
            spin.blockSignals(True)
            spin.setMinimum(min_value)
            spin.setMaximum(max_value)
            spin.blockSignals(False)

    @Slot(float)
    def _on_spin_frequency_from_edited(self, value: float) -> None:
        self._frequency_from = self._settings.to_mhz(value)
        if not self.signalsBlocked():
            self.frequencyLimitsChanged.emit()

    @Slot(float)
    def _on_spin_frequency_to_edited(self, value: float) -> None:
        self._frequency_to = self._settings.to_mhz(value)
        if not self.signalsBlocked():
            self.frequencyLimitsChanged.emit()

    @Slot(float)
    def _on_spin_frequency_center_edited(self, value: float) -> None:
        self._frequency_center = self._settings.to_mhz(value)
        if not self.signalsBlocked():
            self.frequencyLimitsChanged.emit()

    @Slot(float)
    def _on_spin_frequency_deviation_edited(self, value: float) -> None:
        self._frequency_deviation = self._settings.to_mhz(value)
        if not self.signalsBlocked():
            self.frequencyLimitsChanged.emit()

    def fill_parameters(self) -> None:
        frequency_suffix: int = self._settings.frequency_unit
        frequency_suffix_str: str = " " + self._settings.FREQUENCY_UNITS[frequency_suffix]
        from_mhz: Callable[[float], float] = self._settings.from_mhz
        if frequency_suffix in (0, 1, 2):  # MHz, GHz, cm⁻¹
            self._spin_frequency_from.setValue(from_mhz(self._frequency_from))
            self._spin_frequency_to.setValue(from_mhz(self._frequency_to))
            self._spin_frequency_center.setValue(from_mhz(self._frequency_center))
            self._spin_frequency_deviation.setValue(from_mhz(self._frequency_deviation))
        elif frequency_suffix == 3:  # nm
            self._spin_frequency_from.setValue(from_mhz(self._frequency_from))
            self._spin_frequency_to.setValue(from_mhz(self._frequency_to))
            self._spin_frequency_center.setValue(from_mhz(self._frequency_center))
            self._spin_frequency_deviation.setValue(
                abs(
                    from_mhz(self._frequency_center - self._frequency_deviation)
                    - from_mhz(self._frequency_center + self._frequency_deviation)
                )
                / 2.0
            )
        else:
            raise IndexError("Wrong frequency unit index", frequency_suffix)
        precision: int = [4, 7, 8, 8][frequency_suffix]
        step_factor: float = [2.5, 2.5, 2.5, 0.25][frequency_suffix]
        frequency_spins: list[QDoubleSpinBox] = [
            self._spin_frequency_from,
            self._spin_frequency_to,
            self._spin_frequency_center,
            self._spin_frequency_deviation,
        ]
        for spin in frequency_spins:
            spin.setSuffix(frequency_suffix_str)
            spin.setDecimals(precision)
            spin.setSingleStep(step_factor * self._spin_frequency_deviation.value())



from contextlib import suppress
from typing import Any

from qtpy.QtGui import QIcon, QKeySequence
from qtpy.QtWidgets import QAction, QMenu, QMenuBar, QStyle, QWidget

__all__ = ["MenuBar"]


class MenuBar(QMenuBar):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.menu_file: QMenu = self.addMenu(self.tr("&File"))
        self.action_load: QAction = self.menu_file.addAction(
            self._icon("document-open", "mdi6.folder-open", standard_pixmap=QStyle.StandardPixmap.SP_DialogOpenButton),
            self.tr("&Load Catalog…"),
            QKeySequence.StandardKey.Open,
        )
        self.action_reload: QAction = self.menu_file.addAction(
            self._icon("document-revert", "mdi6.reload", standard_pixmap=QStyle.StandardPixmap.SP_BrowserReload),
            self.tr("&Reload Catalogs"),
            QKeySequence.StandardKey.Refresh,
        )
        self.action_save_as: QAction = self.menu_file.addAction(
            self._icon(
                "document-save-as", "mdi6.content-save-edit", standard_pixmap=QStyle.StandardPixmap.SP_DialogSaveButton
            ),
            self.tr("&Save Catalog As…"),
            QKeySequence.StandardKey.SaveAs,
        )
        self.menu_file.addSeparator()
        self.action_download_catalog: QAction = self.menu_file.addAction(
            self._icon("network-receive", "mdi6.download"), self.tr("&Download Catalog…")
        )
        self.menu_file.addSeparator()
        self.action_preferences: QAction = self.menu_file.addAction(
            self._icon("preferences-other", "mdi6.application-settings"),
            self.tr("&Preferences…"),
            QKeySequence.StandardKey.Preferences,
        )
        self.menu_file.addSeparator()
        self.action_quit: QAction = self.menu_file.addAction(
            self._icon("application-exit", "mdi6.exit-run", standard_pixmap=QStyle.StandardPixmap.SP_DialogCloseButton),
            self.tr("&Quit"),
            QKeySequence.StandardKey.Quit,
        )

        self.menu_edit: QMenu = self.addMenu(self.tr("&Edit"))
        self.action_clear: QAction = self.menu_edit.addAction(
            self._icon("edit-clear", "mdi6.broom", standard_pixmap=QStyle.StandardPixmap.SP_DialogResetButton),
            self.tr("Clea&r Results"),
        )
        self.menu_edit.addSeparator()
        self.menu_copy_only: QMenu = self.menu_edit.addMenu(self.tr("Copy &Only"))
        self.action_copy_current: QAction = self.menu_copy_only.addAction(self.tr("Active &Cell"), "Ctrl+Shift+C")
        self.menu_copy_only.addSeparator()
        self.action_copy_name: QAction = self.menu_copy_only.addAction(self.tr("&Substance Name"), "Ctrl+Shift+C, N")
        self.action_copy_frequency: QAction = self.menu_copy_only.addAction(self.tr("&Frequency"), "Ctrl+Shift+C, F")
        self.action_copy_intensity: QAction = self.menu_copy_only.addAction(self.tr("&Intensity"), "Ctrl+Shift+C, I")
        self.action_copy_lower_state_energy: QAction = self.menu_copy_only.addAction(
            self.tr("&Lower State Energy"), "Ctrl+Shift+C, E"
        )
        self.action_copy: QAction = self.menu_edit.addAction(
            self._icon("edit-copy", "mdi6.content-copy"), self.tr("&Copy Selection"), QKeySequence.StandardKey.Copy
        )
        self.menu_edit.addSeparator()
        self.action_select_all: QAction = self.menu_edit.addAction(
            self._icon("edit-select-all", "mdi6.select-all"), self.tr("Select &All"), QKeySequence.StandardKey.SelectAll
        )
        self.menu_edit.addSeparator()
        self.action_substance_info: QAction = self.menu_edit.addAction(
            self._icon(
                "dialog-information",
                "mdi6.flask-empty-outline",
                "mdi6.information-variant",
                options=[{}, {"scale_factor": 0.5}],
            ),
            self.tr("Substance &Info"),
            "Ctrl+I",
        )

        self.menu_columns: QMenu = self.addMenu(self.tr("&Columns"))
        self.action_show_substance: QAction = self.menu_columns.addAction(self.tr("&Substance"))
        self.action_show_frequency: QAction = self.menu_columns.addAction(self.tr("&Frequency"))
        self.action_show_intensity: QAction = self.menu_columns.addAction(self.tr("&Intensity"))
        self.action_show_lower_state_energy: QAction = self.menu_columns.addAction(self.tr("&Lower State Energy"))

        self.menu_help: QMenu = self.addMenu(self.tr("&Help"))
        self.action_check_updates: QAction = self.menu_help.addAction(
            self._icon("application-update", "mdi6.update"), self.tr("Check for &Updates…")
        )
        self.menu_help.addSeparator()
        self.action_about_catalogs: QAction = self.menu_help.addAction(
            self._icon("document-properties", "mdi6.information"), self.tr("About &Catalogs…")
        )
        self.action_about: QAction = self.menu_help.addAction(
            self._icon("help-about", "mdi6.help", standard_pixmap=QStyle.StandardPixmap.SP_FileDialogInfoView),
            self.tr("&About…"),
            QKeySequence.StandardKey.HelpContents,
        )
        self.action_about_qt: QAction = self.menu_help.addAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TitleBarMenuButton), self.tr("About &Qt…")
        )

        self.action_preferences.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.action_quit.setMenuRole(QAction.MenuRole.QuitRole)
        self.action_about.setMenuRole(QAction.MenuRole.AboutRole)
        self.action_about_qt.setMenuRole(QAction.MenuRole.AboutQtRole)

        self.action_save_as.setDisabled(True)

        self.action_show_substance.setCheckable(True)
        self.action_show_frequency.setCheckable(True)
        self.action_show_intensity.setCheckable(True)
        self.action_show_lower_state_energy.setCheckable(True)

    def _icon(
        self, theme_name: str, *qta_name: str, standard_pixmap: QStyle.StandardPixmap | None = None, **qta_specs: Any
    ) -> QIcon:
        if theme_name and QIcon.hasThemeIcon(theme_name):
            return QIcon.fromTheme(theme_name)

        if qta_name:
            with suppress(ImportError, Exception):
                import qtawesome as qta

                return qta.icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

        if standard_pixmap is not None:
            return self.style().standardIcon(standard_pixmap)

        return QIcon()



#!/usr/bin/env python3


def _third_party_modules() -> list[str]:
    import site
    import sys

    prefixes: list[str] = site.getsitepackages([sys.exec_prefix, sys.prefix])
    third_party_modules: list[str] = []
    for module_name, module in sys.modules.copy().items():
        paths = getattr(module, "__path__", [])
        if (
            "." not in module_name
            and module_name != "_distutils_hack"
            and paths
            and getattr(module, "__package__", "")
            and any(p.startswith(prefix) for p in paths for prefix in prefixes)
        ):
            third_party_modules.append(module_name)

    return third_party_modules


def test_cli():
    import sys
    from importlib.util import find_spec

    from pycatsearch import main_cli as main

    third_party_modules: list[str]

    third_party_modules = _third_party_modules()
    assert third_party_modules == [], third_party_modules

    assert main() != 0

    sys.argv.append("catalog.json.gz")
    assert main() != 0

    sys.argv.extend("--min-frequency 118749 --max-frequency 118751 -n oxygen".split())
    assert main() == 0

    third_party_modules = _third_party_modules()
    assert third_party_modules == ["orjson"] * (find_spec("orjson") is not None), third_party_modules


if __name__ == "__main__":
    import sys
    from os import path

    sys.path = list(set(sys.path) | {path.abspath(path.join(__file__, path.pardir, path.pardir))})

    test_cli()



from qtpy.QtCore import QModelIndex, QPersistentModelIndex, QRect, QSize
from qtpy.QtGui import QAbstractTextDocumentLayout, QPainter, QPalette, QTextDocument
from qtpy.QtWidgets import QApplication, QStyle, QStyleOptionViewItem, QStyledItemDelegate

__all__ = ["HTMLDelegate"]


class HTMLDelegate(QStyledItemDelegate):
    def paint(
        self,
        painter: QPainter,
        option: QStyleOptionViewItem,
        index: QModelIndex | QPersistentModelIndex,
    ) -> None:
        self.initStyleOption(option, index)
        style: QStyle
        if option.widget:
            style = option.widget.style()
        else:
            style = QApplication.style()
        doc: QTextDocument = QTextDocument()
        doc.setHtml(option.text)
        option.text = ""
        style.drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter)
        ctx: QAbstractTextDocumentLayout.PaintContext = QAbstractTextDocumentLayout.PaintContext()
        if option.state & QStyle.StateFlag.State_Selected:
            ctx.palette.setBrush(QPalette.ColorRole.Text, option.palette.highlightedText())
        text_rect: QRect = style.subElementRect(QStyle.SubElement.SE_ItemViewItemText, option)
        painter.save()
        painter.translate(text_rect.topLeft())
        painter.setClipRect(option.rect.translated(-text_rect.topLeft()))
        painter.translate(0, 0.5 * (option.rect.height() - doc.size().height()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option: QStyleOptionViewItem | None, index: QModelIndex | QPersistentModelIndex) -> QSize:
        options: QStyleOptionViewItem = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        doc: QTextDocument = QTextDocument()
        doc.setHtml(options.text)
        doc.setTextWidth(options.rect.width())
        return QSize(round(doc.idealWidth()), round(QTextDocument().size().height()))



from typing import cast

from qtpy.QtCore import QLocale
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget, QWizard, QWizardPage

__all__ = ["SummaryPage"]


class SummaryPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        layout: QVBoxLayout = QVBoxLayout(self)
        self._label: QLabel = QLabel(self)
        layout.addWidget(self._label)

    def initializePage(self) -> None:
        from . import SaveCatalogWizard

        super(SummaryPage, self).initializePage()
        if cast(SaveCatalogWizard, self.wizard()).catalog:
            self.setTitle(self.tr("Success"))
            self.setButtonText(QWizard.WizardButton.FinishButton, self.tr("&Save"))
            self._label.setText(
                self.tr("Click {button_text} to save the catalog for {min_frequency} to {max_frequency} MHz.").format(
                    button_text=self.buttonText(QWizard.WizardButton.FinishButton).replace("&", ""),
                    min_frequency=QLocale().toString(self.field("min_frequency")),
                    max_frequency=QLocale().toString(self.field("max_frequency")),
                )
            )
        else:
            self.setTitle(self.tr("Failure"))
            self._label.setText(self.tr("For the specified frequency range, nothing has been loaded."))



import enum
from typing import Any, Callable, Final

from qtpy.QtCore import QAbstractTableModel, QLocale, QModelIndex, QPersistentModelIndex, Qt
from qtpy.QtWidgets import QWidget

from .settings import Settings
from ..utils import CatalogType, best_name

__all__ = ["FoundLinesModel"]


class FoundLinesModel(QAbstractTableModel):
    ROW_BATCH_COUNT: Final[int] = 5

    class DataType:
        __slots__ = [
            "precision",
            "decimal_point",
            "species_tag",
            "name",
            "frequency",
            "intensity",
            "lower_state_energy",
            "_frequency_str",
            "_intensity_str",
            "_lower_state_energy_str",
        ]

        def __init__(
            self,
            precision: int,
            decimal_point: str,
            species_tag: int,
            name: str,
            frequency: float,
            intensity: float,
            lower_state_energy: float,
        ) -> None:
            self.precision: int = precision
            self.decimal_point: str = decimal_point

            self.species_tag: int = species_tag
            self.name: str = name
            self.frequency: float = frequency
            self.intensity: float = intensity
            self.lower_state_energy: float = lower_state_energy

            self._frequency_str: str | None = None
            self._intensity_str: str | None = None
            self._lower_state_energy_str: str | None = None

        @property
        def frequency_str(self) -> str:
            if self._frequency_str is None:
                self._frequency_str = f"{self.frequency:.{self.precision}f}".replace(".", self.decimal_point)
            return self._frequency_str

        @property
        def intensity_str(self) -> str:
            if self._intensity_str is None:
                if self.intensity == 0.0:
                    self._intensity_str = "0"
                elif abs(self.intensity) < 0.1:
                    self._intensity_str = f"{self.intensity:.4e}".replace(".", self.decimal_point)
                else:
                    self._intensity_str = f"{self.intensity:.4f}".replace(".", self.decimal_point)
            return self._intensity_str

        @property
        def lower_state_energy_str(self) -> str:
            if self._lower_state_energy_str is None:
                if self.lower_state_energy == 0.0:
                    self._lower_state_energy_str = "0"
                elif abs(self.lower_state_energy) < 0.1:
                    self._lower_state_energy_str = f"{self.lower_state_energy:.4e}".replace(".", self.decimal_point)
                else:
                    self._lower_state_energy_str = f"{self.lower_state_energy:.4f}".replace(".", self.decimal_point)
            return self._lower_state_energy_str

        def __eq__(self, other: Any) -> bool:
            if not isinstance(other, FoundLinesModel.DataType):
                raise NotImplementedError(f"Comparison with {type(other)} is not supported")
            return (
                self.species_tag == other.species_tag
                and self.frequency == other.frequency
                and self.intensity == other.intensity
                and self.lower_state_energy == other.lower_state_energy
            )

        def __hash__(self) -> int:
            return hash(self.species_tag) ^ hash(self.frequency) ^ hash(self.lower_state_energy)

    class Columns(enum.IntEnum):
        SubstanceName = 0
        Frequency = 1
        Intensity = 2
        LowerStateEnergy = 3

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._settings: Settings = settings
        self._data: list[FoundLinesModel.DataType] = []
        self._rows_loaded: int = 0

        unit_format: Final[str] = self.tr("{value} [{unit}]", "unit format")
        self._header: Final[list[str]] = [
            self.tr("Substance"),
            unit_format.format(value=self.tr("Frequency"), unit=self._settings.frequency_unit_str),
            unit_format.format(value=self.tr("Intensity"), unit=self._settings.intensity_unit_str),
            unit_format.format(value=self.tr("Lower state energy"), unit=self._settings.energy_unit_str),
        ]

    def update_units(self) -> None:
        unit_format: Final[str] = self.tr("{value} [{unit}]", "unit format")
        self._header[FoundLinesModel.Columns.Frequency] = unit_format.format(
            value=self.tr("Frequency"),
            unit=self._settings.frequency_unit_str,
        )
        self._header[FoundLinesModel.Columns.Intensity] = unit_format.format(
            value=self.tr("Intensity"),
            unit=self._settings.intensity_unit_str,
        )
        self._header[FoundLinesModel.Columns.LowerStateEnergy] = unit_format.format(
            value=self.tr("Lower state energy"),
            unit=self._settings.energy_unit_str,
        )

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return min(len(self._data), self._rows_loaded)

    def columnCount(self, parent: QModelIndex | QPersistentModelIndex = ...) -> int:
        return len(self._header)

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> str | None:
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                item: FoundLinesModel.DataType = self._data[index.row()]
                column_index: int = index.column()
                if column_index == FoundLinesModel.Columns.SubstanceName:
                    return item.name
                if column_index == FoundLinesModel.Columns.Frequency:
                    return item.frequency_str
                if column_index == FoundLinesModel.Columns.Intensity:
                    return item.intensity_str
                if column_index == FoundLinesModel.Columns.LowerStateEnergy:
                    return item.lower_state_energy_str
        return None

    def row(self, row_index: int) -> DataType:
        return self._data[row_index]

    def headerData(self, col: int, orientation: Qt.Orientation, role: int = ...) -> str | None:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[col]
        return None

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value: str, role: int = ...) -> bool:
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.DisplayRole
            and 0 <= section < len(self._header)
        ):
            self._header[section] = value
            return True
        return False

    def clear(self) -> None:
        self.set_entries(dict())

    def set_entries(self, entries: CatalogType) -> None:
        from_mhz: Callable[[float], float] = self._settings.from_mhz
        from_log10_sq_nm_mhz: Callable[[float], float] = self._settings.from_log10_sq_nm_mhz
        from_rec_cm: Callable[[float], float] = self._settings.from_rec_cm
        frequency_suffix: int = self._settings.frequency_unit
        precision: int = [4, 7, 8, 8][frequency_suffix]
        locale: QLocale = QLocale()
        decimal_point: str = locale.decimalPoint()

        self.beginResetModel()
        rich_text_in_formulas: bool = self._settings.rich_text_in_formulas
        self._data = list(
            set(
                FoundLinesModel.DataType(
                    precision=precision,
                    decimal_point=decimal_point,
                    species_tag=species_tag,
                    name=best_name(entries[species_tag], rich_text_in_formulas),
                    frequency=from_mhz(line.frequency),
                    intensity=from_log10_sq_nm_mhz(line.intensity),
                    lower_state_energy=from_rec_cm(line.lowerstateenergy),
                )
                for species_tag in entries
                for line in entries[species_tag].lines
            )
        )
        self._rows_loaded = 0
        self.endResetModel()

    def sort(self, column: int, order: Qt.SortOrder = Qt.SortOrder.AscendingOrder) -> None:
        self.beginResetModel()
        key = {
            FoundLinesModel.Columns.SubstanceName: (
                lambda line: (line.name, line.frequency, line.intensity, line.lower_state_energy)
            ),
            FoundLinesModel.Columns.Frequency: (
                lambda line: (line.frequency, line.intensity, line.name, line.lower_state_energy)
            ),
            FoundLinesModel.Columns.Intensity: (
                lambda line: (line.intensity, line.frequency, line.name, line.lower_state_energy)
            ),
            FoundLinesModel.Columns.LowerStateEnergy: (
                lambda line: (line.lower_state_energy, line.intensity, line.frequency, line.name)
            ),
        }[FoundLinesModel.Columns(column)]
        self._data.sort(key=key, reverse=bool(order != Qt.SortOrder.AscendingOrder))
        self.endResetModel()

    def canFetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> bool:
        return len(self._data) > self._rows_loaded

    def fetchMore(self, index: QModelIndex | QPersistentModelIndex = QModelIndex()) -> None:
        # https://sateeshkumarb.wordpress.com/2012/04/01/paginated-display-of-table-data-in-pyqt/
        remainder: int = len(self._data) - self._rows_loaded
        if remainder <= 0:
            return
        items_to_fetch: int = min(remainder, FoundLinesModel.ROW_BATCH_COUNT)
        self.beginInsertRows(index, self._rows_loaded, self._rows_loaded + items_to_fetch - 1)
        self._rows_loaded += items_to_fetch
        self.endInsertRows()



import site

from setuptools import setup

site.main()


def required_packages() -> list[str]:
    import sys

    if sys.version_info < (3, 8):
        raise SystemError("Python versions prior to 3.8 are not supported")

    import platform
    from typing import Collection, NamedTuple, Sequence

    from packaging.version import Version

    class PackageRequirement(NamedTuple):
        package_name: str
        min_version: str = ""
        blacklisted_versions: Collection[str] = ()

        def __str__(self) -> str:
            return self.package_name + ",".join(
                ([" >= " + self.min_version] if self.min_version else [])
                + [(" != " + blv) for blv in self.blacklisted_versions]
            )

    def is_package_importable(package_requirement: PackageRequirement) -> bool:
        from importlib.metadata import version, PackageNotFoundError

        try:
            actual_package_version: Version = Version(version(package_requirement.package_name))
        except PackageNotFoundError:
            return False
        else:
            if package_requirement.min_version and (actual_package_version < Version(package_requirement.min_version)):
                return False
            if actual_package_version in package_requirement.blacklisted_versions:
                return False
        return True

    def required_package(package_requirement: PackageRequirement | Sequence[PackageRequirement]) -> PackageRequirement:
        """
        Install packages if missing

        :param package_requirement: A package name or a sequence of the alternative packages names;
                             if none of the packages installed beforehand, install the first one given
        :returns bool: `True` if a package is importable, `False` when an attempt to install the package is made
        """

        if not package_requirement:
            raise ValueError("No package requirements given")

        if not isinstance(package_requirement, PackageRequirement) and isinstance(package_requirement, Sequence):
            for _package_requirement in package_requirement:
                if is_package_importable(_package_requirement):
                    return _package_requirement
            return package_requirement[0]

        if isinstance(package_requirement, PackageRequirement):
            return package_requirement
        raise ValueError("Invalid requirements")

    qt_list: list[PackageRequirement]
    uname: platform.uname_result = platform.uname()
    if (
        # Windows 10 21H2 or later required
        uname.system == "Windows" and Version(uname.version) < Version("10.0.19044")
    ) or uname.machine not in ("x86_64", "AMD64"):
        # Qt6 does not support the OSes
        qt_list = [
            PackageRequirement(package_name="PyQt5", min_version="5.15.0"),
            PackageRequirement(package_name="PySide2", min_version="5.15.0"),
        ]
    else:
        qt_list = [
            PackageRequirement(
                package_name="PySide6-Essentials",
                min_version="6.2.0",
                blacklisted_versions=("6.8.0", "6.8.0.1"),
            ),
            PackageRequirement(package_name="PyQt6", min_version="6.2.0"),
            PackageRequirement(package_name="PyQt5", min_version="5.15.0"),
            PackageRequirement(package_name="PySide2", min_version="5.15.0"),
        ]

    requirements: list[PackageRequirement | Sequence[PackageRequirement]] = [
        qt_list,
        PackageRequirement(package_name="qtpy", min_version="2.4.1"),
    ]
    return [str(required_package(requirement)) for requirement in requirements]


setup(install_requires=required_packages())



from qtpy.QtWidgets import QWidget

from .selectable_label import SelectableLabel

__all__ = ["URLLabel"]


class URLLabel(SelectableLabel):
    """A label with selectable hyperlink"""

    def __init__(self, url: str, text: str, parent: QWidget | None = None) -> None:
        super().__init__(f'<a href="{url}">{text}</a>', parent)
        self.setOpenExternalLinks(True)
        self.setToolTip(url)



from os import PathLike

from qtpy import QT5
from qtpy.QtWidgets import QWidget

from .waiting_screen import WaitingScreen
from ..catalog import CatalogType
from ..utils import save_catalog_to_file

__all__ = ["SaveCatalogWaitingScreen"]


class SaveCatalogWaitingScreen(WaitingScreen):
    def __init__(
        self,
        parent: QWidget,
        *,
        filename: str | PathLike[str],
        catalog: CatalogType,
        frequency_limits: tuple[float, float],
        margins: float | None = None,
    ) -> None:
        if QT5:
            super(QWidget, self).__init__(parent)
        super().__init__(
            parent,
            label=self.tr("Please wait…"),
            target=save_catalog_to_file,
            kwargs=dict(filename=filename, catalog=catalog, frequency_limits=frequency_limits),
            margins=margins,
        )



from abc import abstractmethod
from functools import partial
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, ClassVar, Hashable, cast

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QFormLayout,
    QListWidget,
    QListWidgetItem,
    QScrollArea,
    QSpinBox,
    QSplitter,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from .open_file_path_entry import OpenFilePathEntry
from .settings import Settings

__all__ = ["Preferences"]


class BaseLogger:
    logger: ClassVar[Logger]
    from typing import ParamSpec

    _P = ParamSpec("_P")

    def __new__(cls, *args: _P.args, **kwargs: _P.kwargs):
        cls.logger = getLogger(cls.__name__)
        return super().__new__(cls)

    @abstractmethod
    def __init__(self, *args: _P.args, **kwargs: _P.kwargs) -> None:
        pass


class PreferencePage(BaseLogger, QScrollArea):
    """A page of the Preferences dialog"""

    def __init__(
        self,
        value: dict[
            (str | tuple[str, tuple[str, ...]] | tuple[str, tuple[str, ...], tuple[tuple[str, Any], ...]]),
            (
                Settings.CallbackOnly
                | Settings.PathCallbackOnly
                | Settings.SpinboxAndCallback
                | Settings.ComboboxAndCallback
                | Settings.EditableComboboxAndCallback
            ),
        ],
        settings: Settings,
        parent: QWidget | None = None,
    ) -> None:
        BaseLogger.__init__(self)
        QScrollArea.__init__(self, parent)

        widget: QWidget = QWidget(self)
        self.setWidget(widget)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setFrameStyle(0)

        self._changed_settings: dict[str, Any] = {}

        # https://forum.qt.io/post/671245
        def _on_event(x: bool | int | float | str, *, callback: str) -> None:
            self._changed_settings[callback] = x

        def _on_combo_box_current_index_changed(_: int, *, sender: QComboBox, callback: str) -> None:
            self._changed_settings[callback] = sender.currentData()

        if not (isinstance(value, dict) and value):
            raise TypeError(f"Invalid type: {type(value)}")
        layout: QFormLayout = QFormLayout(widget)
        layout.setLabelAlignment(layout.labelAlignment() | Qt.AlignmentFlag.AlignVCenter)
        key2: str
        value2: (
            Settings.CallbackOnly
            | Settings.PathCallbackOnly
            | Settings.SpinboxAndCallback
            | Settings.ComboboxAndCallback
            | Settings.EditableComboboxAndCallback
        )

        check_box: QCheckBox
        path_entry: OpenFilePathEntry
        spin_box: QSpinBox | QDoubleSpinBox
        combo_box: QComboBox

        for key2, value2 in value.items():
            current_value: Any = getattr(settings, value2.callback)
            if isinstance(value2, Settings.CallbackOnly):
                if isinstance(current_value, bool):
                    check_box = QCheckBox(settings.tr(key2), widget)
                    check_box.setChecked(current_value)
                    check_box.toggled.connect(partial(_on_event, callback=value2.callback))
                    layout.addWidget(check_box)
                elif isinstance(current_value, Path):
                    path_entry = OpenFilePathEntry(current_value, widget)
                    path_entry.changed.connect(partial(_on_event, callback=value2.callback))
                    layout.addRow(key2, path_entry)
                else:
                    PreferencePage.logger.error(f"The type of {value2.callback!r} is not supported")
            elif isinstance(value2, Settings.PathCallbackOnly):
                if isinstance(current_value, (Path, type(None))):
                    path_entry = OpenFilePathEntry(current_value, widget)
                    path_entry.changed.connect(partial(_on_event, callback=value2.callback))
                    layout.addRow(key2, path_entry)
                else:
                    PreferencePage.logger.error(f"The type of {value2.callback!r} is not supported")
            elif isinstance(value2, Settings.SpinboxAndCallback):
                if isinstance(current_value, int):
                    spin_box = QSpinBox(widget)
                elif isinstance(current_value, float):
                    spin_box = QDoubleSpinBox(widget)
                else:
                    PreferencePage.logger.error(f"The type of {value2.callback!r} is not supported")
                    continue
                spin_box.setRange(value2.range.start, value2.range.stop)
                spin_box.setSingleStep(value2.range.step or 1)
                spin_box.setValue(current_value)
                spin_box.setPrefix(value2.prefix_and_suffix[0])
                spin_box.setSuffix(value2.prefix_and_suffix[1])
                spin_box.valueChanged.connect(partial(_on_event, callback=value2.callback))
                layout.addRow(key2, spin_box)
            elif isinstance(value2, Settings.ComboboxAndCallback):
                if not isinstance(current_value, Hashable):
                    PreferencePage.logger.error(f"The type of {value2.callback!r} is not supported")
                    continue
                combo_box = QComboBox(widget)
                combobox_data: dict[Hashable, str]
                if isinstance(value2.combobox_data, dict):
                    combobox_data = value2.combobox_data
                else:
                    combobox_data = dict(enumerate(value2.combobox_data))
                for index, (data, item) in enumerate(combobox_data.items()):
                    combo_box.addItem(settings.tr(item), data)
                combo_box.setEditable(False)
                combo_box.setCurrentText(combobox_data[current_value])
                combo_box.currentIndexChanged.connect(
                    partial(
                        _on_combo_box_current_index_changed,
                        sender=combo_box,
                        callback=value2.callback,
                    )
                )
                layout.addRow(key2, combo_box)
            elif isinstance(value2, Settings.EditableComboboxAndCallback):
                if isinstance(current_value, str):
                    current_text: str = current_value
                else:
                    PreferencePage.logger.error(f"The type of {value2.callback!r} is not supported")
                    continue
                combo_box = QComboBox(widget)
                combo_box.addItems(value2.combobox_items)
                if current_text in value2.combobox_items:
                    combo_box.setCurrentIndex(value2.combobox_items.index(current_text))
                else:
                    combo_box.insertItem(0, current_text)
                    combo_box.setCurrentIndex(0)
                combo_box.setEditable(True)
                combo_box.currentTextChanged.connect(partial(_on_event, callback=value2.callback))
                layout.addRow(key2, combo_box)
            else:
                PreferencePage.logger.error(f"{value2!r} is not supported")

    @property
    def changed_settings(self) -> dict[str, Any]:
        return self._changed_settings.copy()


class PreferencesBody(BaseLogger, QSplitter):
    """The main area of the GUI preferences dialog"""

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        from . import qta_icon  # import locally to avoid a circular import

        BaseLogger.__init__(self)
        QSplitter.__init__(self, parent)
        self.setObjectName("preferencesBody")

        self.setOrientation(Qt.Orientation.Horizontal)
        self.setChildrenCollapsible(False)
        content: QListWidget = QListWidget(self)
        self._stack: QStackedWidget = QStackedWidget(self)
        key: str | tuple[str, tuple[str, ...]] | tuple[str, tuple[str, ...], tuple[tuple[str, Any], ...]]
        value: dict[
            str,
            (
                Settings.CallbackOnly
                | Settings.PathCallbackOnly
                | Settings.SpinboxAndCallback
                | Settings.ComboboxAndCallback
                | Settings.EditableComboboxAndCallback
            ),
        ]
        for key, value in settings.dialog.items():
            if not (isinstance(value, dict) and value):
                PreferencesBody.logger.error(f"Invalid value of {key!r}")
                continue
            new_item: QListWidgetItem
            if isinstance(key, str):
                new_item = QListWidgetItem(key)
            elif isinstance(key, tuple):
                if len(key) == 1:
                    new_item = QListWidgetItem(key[0])
                elif len(key) == 2:
                    new_item = QListWidgetItem(qta_icon(*key[1]), key[0])
                elif len(key) == 3:
                    new_item = QListWidgetItem(qta_icon(*key[1], **dict(key[2])), key[0])
                else:
                    PreferencesBody.logger.error(f"Invalid key: {key!r}")
                    continue
            else:
                PreferencesBody.logger.error(f"Invalid key type: {key!r}")
                continue
            content.addItem(new_item)
            box: PreferencePage = PreferencePage(value, settings, self._stack)
            self._stack.addWidget(box)
        content.setMinimumWidth(content.sizeHintForColumn(0) + 2 * content.frameWidth())
        self.addWidget(content)
        self.addWidget(self._stack)

        if content.count() > 0:
            content.setCurrentRow(0)  # select the first page

        content.currentRowChanged.connect(self._stack.setCurrentIndex)

    @property
    def changed_settings(self) -> dict[str, Any]:
        changed_settings: dict[str, Any] = {}
        for index in range(self._stack.count()):
            changed_settings.update(cast(PreferencePage, self._stack.widget(index)).changed_settings)
        return changed_settings


class Preferences(QDialog):
    """GUI preferences dialog"""

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setObjectName("preferencesDialog")

        self._settings: Settings = settings
        self.setModal(True)
        self.setWindowTitle(self.tr("Preferences"))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())

        layout: QVBoxLayout = QVBoxLayout(self)
        self._preferences_body: PreferencesBody = PreferencesBody(settings=settings, parent=parent)
        layout.addWidget(self._preferences_body)
        buttons: QDialogButtonBox = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel,
            self,
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.close)
        layout.addWidget(buttons)

        self.adjustSize()
        self.resize(self.width() + 4, self.height())

        self._settings.restore(self)
        self._settings.restore(self._preferences_body)

    def reject(self) -> None:
        self._settings.save(self)
        self._settings.save(self._preferences_body)
        return super().reject()

    def accept(self) -> None:
        self._settings.save(self)
        self._settings.save(self._preferences_body)

        for key, value in self._preferences_body.changed_settings.items():
            setattr(self._settings, key, value)
        return super().accept()



from contextlib import contextmanager, suppress
from os import PathLike, linesep
from pathlib import Path
from typing import Any, Callable, Final, Hashable, Iterable, NamedTuple, Sequence

from qtpy.QtCore import QByteArray, QObject, QSettings

from ..utils import (
    cm_per_molecule_to_log10_sq_nm_mhz,
    ghz_to_mhz,
    j_to_rec_cm,
    log10_cm_per_molecule_to_log10_sq_nm_mhz,
    log10_sq_nm_mhz_to_cm_per_molecule,
    log10_sq_nm_mhz_to_log10_cm_per_molecule,
    log10_sq_nm_mhz_to_sq_nm_mhz,
    meV_to_rec_cm,
    mhz_to_ghz,
    mhz_to_nm,
    mhz_to_rec_cm,
    nm_to_mhz,
    rec_cm_to_j,
    rec_cm_to_meV,
    rec_cm_to_mhz,
    sq_nm_mhz_to_log10_sq_nm_mhz,
)

__all__ = ["Settings"]


class Settings(QSettings):
    """convenient internal representation of the application settings"""

    class CallbackOnly(NamedTuple):
        callback: str

    class PathCallbackOnly(NamedTuple):
        callback: str

    class SpinboxAndCallback(NamedTuple):
        range: slice
        prefix_and_suffix: tuple[str, str]
        callback: str

    class ComboboxAndCallback(NamedTuple):
        combobox_data: Iterable[str] | dict[Hashable, str]
        callback: str

    class EditableComboboxAndCallback(NamedTuple):
        combobox_items: Sequence[str]
        callback: str

    TO_MHZ: Final[list[Callable[[float], float]]] = [lambda x: x, ghz_to_mhz, rec_cm_to_mhz, nm_to_mhz]
    FROM_MHZ: Final[list[Callable[[float], float]]] = [lambda x: x, mhz_to_ghz, mhz_to_rec_cm, mhz_to_nm]

    TO_LOG10_SQ_NM_MHZ: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        sq_nm_mhz_to_log10_sq_nm_mhz,
        log10_cm_per_molecule_to_log10_sq_nm_mhz,
        cm_per_molecule_to_log10_sq_nm_mhz,
    ]
    FROM_LOG10_SQ_NM_MHZ: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        log10_sq_nm_mhz_to_sq_nm_mhz,
        log10_sq_nm_mhz_to_log10_cm_per_molecule,
        log10_sq_nm_mhz_to_cm_per_molecule,
    ]

    TO_REC_CM: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        meV_to_rec_cm,
        j_to_rec_cm,
    ]
    FROM_REC_CM: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        rec_cm_to_meV,
        rec_cm_to_j,
    ]

    TO_K: Final[list[Callable[[float], float]]] = [lambda x: x, lambda x: x + 273.15]
    FROM_K: Final[list[Callable[[float], float]]] = [lambda x: x, lambda x: x - 273.15]

    INCHI_KEY_SEARCH_PROVIDERS: Final[list[str]] = [
        "https://pubchem.ncbi.nlm.nih.gov/#query={InChIKey}",
        "https://www.ebi.ac.uk/unichem/compoundsources?type=inchikey&compound={InChIKey}",
        "https://webbook.nist.gov/cgi/cbook.cgi?InChI={InChIKey}",
        "https://www.spectrabase.com/search?q={InChIKey}",
        "https://www.google.com/search?q={InChIKey}",
        "http://gmd.mpimp-golm.mpg.de/search.aspx?query={InChIKey}",
        "http://www.chemspider.com/InChIKey/{InChIKey}",
    ]

    def __init__(self, organization: str, application: str, parent: QObject | None = None) -> None:
        super().__init__(organization, application, parent)

        # for some reason, the dicts are not being translated when used as class variables
        self.LINE_ENDS: Final[dict[str, str]] = {
            "\n": self.tr(r"Line Feed (\n)"),
            "\r": self.tr(r"Carriage Return (\r)"),
            "\r\n": self.tr(r"CR+LF (\r\n)"),
            "\n\r": self.tr(r"LF+CR (\n\r)"),
        }
        self.CSV_SEPARATORS: Final[dict[str, str]] = {
            ",": self.tr(r"comma (,)"),
            "\t": self.tr(r"tab (\t)"),
            ";": self.tr(r"semicolon (;)"),
            " ": self.tr(r"space ( )"),
        }
        self.FREQUENCY_UNITS: Final[list[str]] = [self.tr("MHz"), self.tr("GHz"), self.tr("cm⁻¹"), self.tr("nm")]
        self.INTENSITY_UNITS: Final[list[str]] = [
            self.tr("lg(nm² × MHz)"),
            self.tr("nm² × MHz"),
            self.tr("lg(cm / molecule)"),
            self.tr("cm / molecule"),
        ]
        self.ENERGY_UNITS: Final[list[str]] = [self.tr("cm⁻¹"), self.tr("meV"), self.tr("J")]
        self.TEMPERATURE_UNITS: Final[list[str]] = [self.tr("K"), self.tr("°C")]

    @property
    def dialog(
        self,
    ) -> dict[
        (str | tuple[str, tuple[str, ...]] | tuple[str, tuple[str, ...], tuple[tuple[str, Any], ...]]),
        dict[str, (CallbackOnly | PathCallbackOnly | SpinboxAndCallback | ComboboxAndCallback)],
    ]:
        return {
            (self.tr("When the program starts"), ("mdi6.rocket-launch",)): {
                self.tr("Load catalogs"): Settings.CallbackOnly(Settings.load_last_catalogs.fset.__name__),
                self.tr("Check for update"): Settings.CallbackOnly(Settings.check_updates.fset.__name__),
            },
            (self.tr("Display"), ("mdi6.binoculars",)): {
                self.tr("Allow rich text in formulas"): Settings.CallbackOnly(
                    Settings.rich_text_in_formulas.fset.__name__
                ),
            },
            (self.tr("Units"), ("mdi6.pencil-ruler",)): {
                self.tr("Frequency:"): Settings.ComboboxAndCallback(
                    self.FREQUENCY_UNITS, Settings.frequency_unit.fset.__name__
                ),
                self.tr("Intensity:"): Settings.ComboboxAndCallback(
                    self.INTENSITY_UNITS, Settings.intensity_unit.fset.__name__
                ),
                self.tr("Energy:"): Settings.ComboboxAndCallback(self.ENERGY_UNITS, Settings.energy_unit.fset.__name__),
                self.tr("Temperature:"): Settings.ComboboxAndCallback(
                    self.TEMPERATURE_UNITS, Settings.temperature_unit.fset.__name__
                ),
            },
            (self.tr("Export"), ("mdi6.file-export",)): {
                self.tr("With units"): Settings.CallbackOnly(Settings.with_units.fset.__name__),
                self.tr("Line ending:"): Settings.ComboboxAndCallback(self.LINE_ENDS, Settings.line_end.fset.__name__),
                self.tr("CSV separator:"): Settings.ComboboxAndCallback(
                    self.CSV_SEPARATORS, Settings.csv_separator.fset.__name__
                ),
            },
            (
                self.tr("Info"),
                ("mdi6.flask-empty-outline", "mdi6.information-variant"),
                (("options", ((), (("scale_factor", 0.5),))),),
            ): {
                self.tr("InChI key search URL:"): Settings.EditableComboboxAndCallback(
                    self.INCHI_KEY_SEARCH_PROVIDERS, Settings.inchi_key_search_url_template.fset.__name__
                ),
            },
        }

    @contextmanager
    def section(self, section: str) -> None:
        try:
            self.beginGroup(section)
            yield None
        finally:
            self.endGroup()

    @property
    def frequency_unit(self) -> int:
        with self.section("frequency"):
            return self.value("unit", 0, int)

    @frequency_unit.setter
    def frequency_unit(self, new_value: int | str) -> None:
        if isinstance(new_value, str):
            new_value = self.FREQUENCY_UNITS.index(new_value)
        with self.section("frequency"):
            self.setValue("unit", new_value)

    @property
    def frequency_unit_str(self) -> str:
        with self.section("frequency"):
            return self.FREQUENCY_UNITS[self.value("unit", 0, int)]

    @property
    def to_mhz(self) -> Callable[[float], float]:
        with self.section("frequency"):
            return self.TO_MHZ[self.value("unit", 0, int)]

    @property
    def from_mhz(self) -> Callable[[float], float]:
        with self.section("frequency"):
            return self.FROM_MHZ[self.value("unit", 0, int)]

    @property
    def intensity_unit(self) -> int:
        with self.section("intensity"):
            return self.value("unit", 0, int)

    @intensity_unit.setter
    def intensity_unit(self, new_value: int | str) -> None:
        if isinstance(new_value, str):
            new_value = self.INTENSITY_UNITS.index(new_value)
        with self.section("intensity"):
            self.setValue("unit", new_value)

    @property
    def intensity_unit_str(self) -> str:
        with self.section("intensity"):
            return self.INTENSITY_UNITS[self.value("unit", 0, int)]

    @property
    def to_log10_sq_nm_mhz(self) -> Callable[[float], float]:
        with self.section("intensity"):
            return self.TO_LOG10_SQ_NM_MHZ[self.value("unit", 0, int)]

    @property
    def from_log10_sq_nm_mhz(self) -> Callable[[float], float]:
        with self.section("intensity"):
            return self.FROM_LOG10_SQ_NM_MHZ[self.value("unit", 0, int)]

    @property
    def energy_unit(self) -> int:
        with self.section("energy"):
            return self.value("unit", 0, int)

    @energy_unit.setter
    def energy_unit(self, new_value: int | str) -> None:
        if isinstance(new_value, str):
            new_value = self.ENERGY_UNITS.index(new_value)
        with self.section("energy"):
            self.setValue("unit", new_value)

    @property
    def energy_unit_str(self) -> str:
        with self.section("energy"):
            return self.ENERGY_UNITS[self.value("unit", 0, int)]

    @property
    def to_rec_cm(self) -> Callable[[float], float]:
        with self.section("energy"):
            return self.TO_REC_CM[self.value("unit", 0, int)]

    @property
    def from_rec_cm(self) -> Callable[[float], float]:
        with self.section("energy"):
            return self.FROM_REC_CM[self.value("unit", 0, int)]

    @property
    def temperature_unit(self) -> int:
        with self.section("temperature"):
            return self.value("unit", 0, int)

    @temperature_unit.setter
    def temperature_unit(self, new_value: int | str) -> None:
        if isinstance(new_value, str):
            new_value = self.TEMPERATURE_UNITS.index(new_value)
        with self.section("temperature"):
            self.setValue("unit", new_value)

    @property
    def temperature_unit_str(self) -> str:
        with self.section("temperature"):
            return self.TEMPERATURE_UNITS[self.value("unit", 0, int)]

    @property
    def to_k(self) -> Callable[[float], float]:
        with self.section("temperature"):
            return self.TO_K[self.value("unit", 0, int)]

    @property
    def from_k(self) -> Callable[[float], float]:
        with self.section("temperature"):
            return self.FROM_K[self.value("unit", 0, int)]

    @property
    def load_last_catalogs(self) -> bool:
        with self.section("start"):
            return self.value("loadLastCatalogs", True, bool)

    @load_last_catalogs.setter
    def load_last_catalogs(self, new_value: bool) -> None:
        with self.section("start"):
            self.setValue("loadLastCatalogs", new_value)

    @property
    def check_updates(self) -> bool:
        with self.section("start"):
            return self.value("checkUpdates", True, bool)

    @check_updates.setter
    def check_updates(self, new_value: bool) -> None:
        with self.section("start"):
            self.setValue("checkUpdates", new_value)

    @property
    def rich_text_in_formulas(self) -> bool:
        with self.section("display"):
            return self.value("richTextInFormulas", True, bool)

    @rich_text_in_formulas.setter
    def rich_text_in_formulas(self, new_value: bool) -> None:
        with self.section("display"):
            self.setValue("richTextInFormulas", new_value)

    @property
    def line_end(self) -> str:
        with self.section("export"):
            return list(self.LINE_ENDS.keys())[self.value("lineEnd", list(self.LINE_ENDS.keys()).index(linesep), int)]

    @line_end.setter
    def line_end(self, new_value: str) -> None:
        with self.section("export"):
            self.setValue("lineEnd", list(self.LINE_ENDS.keys()).index(new_value))

    @property
    def csv_separator(self) -> str:
        with self.section("export"):
            return list(self.CSV_SEPARATORS.keys())[
                self.value("csvSeparator", list(self.CSV_SEPARATORS.keys()).index("\t"), int)
            ]

    @csv_separator.setter
    def csv_separator(self, new_value: str) -> None:
        with self.section("export"):
            self.setValue("csvSeparator", list(self.CSV_SEPARATORS.keys()).index(new_value))

    @property
    def with_units(self) -> bool:
        with self.section("export"):
            return self.value("withUnits", True, bool)

    @with_units.setter
    def with_units(self, new_value: bool) -> None:
        with self.section("export"):
            self.setValue("withUnits", new_value)

    @property
    def ignored_version(self) -> str:
        with self.section("update"):
            return self.value("ignoredVersion", "", str)

    @ignored_version.setter
    def ignored_version(self, new_value: str) -> None:
        with self.section("update"):
            self.setValue("ignoredVersion", new_value)

    @property
    def inchi_key_search_url_template(self) -> str:
        with self.section("info"):
            return self.value("InChIKeySearchURLTemplate", "https://pubchem.ncbi.nlm.nih.gov/#query={InChIKey}", str)

    @inchi_key_search_url_template.setter
    def inchi_key_search_url_template(self, new_value: str) -> None:
        with self.section("info"):
            self.setValue("InChIKeySearchURLTemplate", new_value)

    @property
    def catalog_file_names(self) -> list[str | PathLike[str]]:
        catalog_file_names: list[str] = []
        with self.section("search"):
            for i in range(self.beginReadArray("catalogFiles")):
                self.setArrayIndex(i)
                path: str = self.value("path", "", str)
                if path:
                    catalog_file_names.append(path)
            self.endArray()
        return catalog_file_names or ["catalog.json.gz", "catalog.json"]

    @catalog_file_names.setter
    def catalog_file_names(self, filenames: Iterable[str | PathLike[str]]) -> None:
        with self.section("search"):
            self.beginWriteArray("catalogFiles")
            for i, s in enumerate(filenames):
                self.setArrayIndex(i)
                self.setValue("path", str(s))
            self.endArray()

    @property
    def translation_path(self) -> Path | None:
        with self.section("translation"):
            v: str = self.value("filePath", "", str)
        return Path(v) if v else None

    @translation_path.setter
    def translation_path(self, new_value: str | PathLike[str] | None) -> None:
        with self.section("translation"):
            self.setValue("filePath", str(new_value or ""))

    @property
    def opened_file_name(self) -> Path | None:
        with self.section("location"):
            v: str = self.value("open", "", str)
        return Path(v) if v else None

    @opened_file_name.setter
    def opened_file_name(self, filename: str | PathLike[str] | None) -> None:
        with self.section("location"):
            self.setValue("open", str(filename or ""))

    @property
    def saved_file_name(self) -> Path | None:
        with self.section("location"):
            v: str = self.value("save", "", str)
        return Path(v) if v else None

    @saved_file_name.setter
    def saved_file_name(self, filename: str | PathLike[str] | None) -> None:
        with self.section("location"):
            self.setValue("save", str(filename or ""))

    def save(self, o: QObject) -> None:
        name: str = o.objectName()
        if not name:
            raise AttributeError(f"No name given for {o}")
        with suppress(AttributeError), self.section("state"):
            # noinspection PyUnresolvedReferences
            self.setValue(name, o.saveState())
        with suppress(AttributeError), self.section("geometry"):
            # noinspection PyUnresolvedReferences
            self.setValue(name, o.saveGeometry())

    def restore(self, o: QObject) -> None:
        name: str = o.objectName()
        if not name:
            raise AttributeError(f"No name given for {o}")
        with suppress(AttributeError), self.section("state"):
            # noinspection PyUnresolvedReferences
            o.restoreState(self.value(name, QByteArray()))
        with suppress(AttributeError), self.section("geometry"):
            # noinspection PyUnresolvedReferences
            o.restoreGeometry(self.value(name, QByteArray()))



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel, QWidget

__all__ = ["SelectableLabel"]


class SelectableLabel(QLabel):
    """A label with selectable text"""

    def __init__(self, text: str, parent: QWidget | None = None) -> None:
        super().__init__(text, parent)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)



import bz2
import copy
import gzip
import lzma
import math
import sys
from contextlib import contextmanager
from datetime import datetime, timezone
from os import PathLike
from pathlib import Path
from typing import BinaryIO, Callable, Collection, Iterable, Iterator, Mapping, NamedTuple, TextIO, cast

try:
    import orjson as json
except ImportError:
    import json

from .utils import (
    CatalogEntryType,
    CatalogJSONType,
    CatalogType,
    LineType,
    M_LOG10E,
    OldCatalogJSONType,
    T0,
    c,
    h,
    k,
    CATALOG,
    BUILD_TIME,
    LINES,
    FREQUENCY,
    SPECIES_TAG,
    merge_sorted,
    search_sorted,
)

__all__ = ["Catalog", "CatalogSourceInfo"]


def filter_by_frequency_and_intensity(
    catalog_entry: CatalogEntryType,
    *,
    min_frequency: float = 0.0,
    max_frequency: float = math.inf,
    min_intensity: float = -math.inf,
    max_intensity: float = math.inf,
    temperature: float = -math.inf,
) -> CatalogEntryType:
    def intensity(_line: LineType) -> float:
        if catalog_entry.degreesoffreedom >= 0 and temperature > 0.0 and temperature != T0:
            return (
                _line.intensity
                + (
                    (0.5 * catalog_entry.degreesoffreedom + 1.0) * math.log(T0 / temperature)
                    - ((1 / temperature - 1 / T0) * _line.lowerstateenergy * 100.0 * h * c / k)
                )
                / M_LOG10E
            )
        else:
            return _line.intensity

    new_catalog_entry: CatalogEntryType = copy.copy(catalog_entry)
    if new_catalog_entry.lines:
        min_frequency_index: int = (
            search_sorted(min_frequency, new_catalog_entry.lines, key=lambda line: line.frequency) + 1
        )
        max_frequency_index: int = search_sorted(
            max_frequency, new_catalog_entry.lines, key=lambda line: line.frequency, maybe_equal=True
        )
        new_catalog_entry.lines = [
            line
            for line in new_catalog_entry.lines[min_frequency_index : (max_frequency_index + 1)]
            if min_intensity <= intensity(line) <= max_intensity
        ]
    return new_catalog_entry


class CatalogSourceInfo(NamedTuple):
    filename: Path
    build_datetime: datetime | None = None


class CatalogData:
    def __init__(self) -> None:
        self.catalog: CatalogType = dict()
        self.frequency_limits: tuple[tuple[float, float], ...] = ()

    def append(
        self,
        new_catalog: CatalogJSONType | OldCatalogJSONType,
        frequency_limits: Collection[tuple[float, float] | list[float]],
    ) -> None:
        if (
            isinstance(frequency_limits, (list, tuple))
            and len(frequency_limits) == 2
            and all(isinstance(fl, float) or fl is None for fl in frequency_limits)
        ):
            frequency_limits = (frequency_limits,)
        catalog: CatalogType
        if isinstance(new_catalog, list):
            catalog = dict((entry[SPECIES_TAG], CatalogEntryType(**entry)) for entry in new_catalog)
        elif isinstance(new_catalog, dict):
            catalog = dict(
                (int(species_tag_str), CatalogEntryType(**new_catalog[species_tag_str]))
                for species_tag_str in new_catalog
            )
        else:
            raise TypeError("Unsupported data type")

        for species_tag in catalog:
            catalog[species_tag].lines = [LineType(**line) for line in catalog[species_tag].lines]

        def squash_same_species_tag_entries() -> None:
            self.catalog[species_tag].lines = merge_sorted(
                self.catalog[species_tag].lines,
                catalog[species_tag].lines,
                key=lambda line: (line.frequency, line.intensity, line.lowerstateenergy),
            )

        def merge_frequency_tuples(*args: tuple[float, float] | list[float]) -> tuple[tuple[float, float], ...]:
            if not args:
                return tuple()
            ranges: tuple[tuple[float, float], ...] = tuple()
            skip: int = 0
            for i in range(len(args)):
                if skip > 0:
                    skip -= 1
                    continue
                current_range: tuple[float, float] = (
                    float(args[i][0]) if args[i][0] is not None else -math.inf,
                    float(args[i][-1]) if args[i][-1] is not None else math.inf,
                )
                current_min: float = min(current_range)
                current_max: float = max(current_range)
                for r in args[1 + i :]:
                    if current_min <= min(r) <= current_max:
                        current_max = max(current_max, *r)
                        skip += 1
                ranges += ((current_min, current_max),)
            return ranges

        if not self.catalog:
            self.catalog = catalog
        else:
            species_tag: int
            for species_tag in catalog:
                if species_tag not in self.catalog:
                    self.catalog[species_tag] = catalog[species_tag]
                else:
                    squash_same_species_tag_entries()
        self.frequency_limits = merge_frequency_tuples(*self.frequency_limits, *frequency_limits)


class Catalog:
    DEFAULT_SUFFIX: str = ".json.gz"

    class Opener:
        OPENERS_BY_SUFFIX: dict[str, Callable] = {
            ".json": open,
            ".json.gz": gzip.open,
            ".json.bz2": bz2.open,
            ".json.xz": lzma.open,
            ".json.lzma": lzma.open,
        }

        OPENERS_BY_SIGNATURE: dict[str, Callable] = {
            b"{": open,
            b"\x1f\x8b": gzip.open,
            b"BZh": bz2.open,
            b"\xfd\x37\x7a\x58\x5a\x00": lzma.open,
        }

        def __init__(self, path: str | PathLike[str]) -> None:
            self._path: Path = Path(path)
            self._opener: Callable
            suffix: str = ""
            s: str
            for s in reversed(self._path.suffixes):
                suffix = s + suffix
                if suffix in Catalog.Opener.OPENERS_BY_SUFFIX:
                    self._opener = Catalog.Opener.OPENERS_BY_SUFFIX[suffix]
                    return
            if self._path.exists():
                max_signature_length: int = max(map(len, Catalog.Opener.OPENERS_BY_SIGNATURE.keys()))
                f: BinaryIO
                with self._path.open("rb") as f:
                    init_bytes: bytes = f.read(max_signature_length)
                key: bytes
                value: Callable
                for key, value in Catalog.Opener.OPENERS_BY_SIGNATURE.items():
                    if init_bytes.startswith(key):
                        self._opener = value
                        return

            raise ValueError(f"Unknown file: {path}")

        @contextmanager
        def open(
            self,
            mode: str,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
        ) -> Iterator[TextIO | BinaryIO]:
            """
            Open a file in a safe way. Create a temporary file when writing.

            See https://stackoverflow.com/a/29491523/8554611, https://stackoverflow.com/a/2333979/8554611
            """
            writing: bool = "w" in mode.casefold()
            if encoding is None and "b" not in mode.casefold():
                encoding = "utf-8"
            tmp_path: Path = self._path.with_name(self._path.name + ".part")

            # manually open and close the file here to close it before replacing if writing
            file: TextIO | BinaryIO = self._opener(
                tmp_path if writing else self._path,
                mode=mode,
                encoding=encoding,
                errors=errors,
                newline=newline,
            )
            try:
                yield file
            finally:
                file.close()
                if writing:
                    tmp_path.replace(self._path)

    def __init__(self, *catalog_file_names: str | PathLike[str]) -> None:
        self._data: CatalogData = CatalogData()
        self._sources: list[CatalogSourceInfo] = []

        filename: Path
        for filename in map(Path, catalog_file_names):
            if filename.exists() and filename.is_file():
                f_in: BinaryIO
                with Catalog.Opener(filename).open("rb") as f_in:
                    content: bytes = f_in.read()
                    try:
                        json_data: dict[str, str | list[float | None] | CatalogJSONType | OldCatalogJSONType] = (
                            json.loads(content)
                        )
                    except (json.JSONDecodeError, UnicodeDecodeError):
                        print(f"{filename} doesn't contain UTF-8 data or is corrupted", file=sys.stderr)
                    else:
                        try:
                            self._data.append(
                                new_catalog=json_data[CATALOG],
                                frequency_limits=json_data.get(FREQUENCY, ((0.0, math.inf),)),
                            )
                            build_datetime: datetime | None = None
                            if BUILD_TIME in json_data:
                                build_datetime = datetime.fromisoformat(cast(str, json_data[BUILD_TIME]))
                        except LookupError:
                            print(f"{filename} is corrupted", file=sys.stderr)
                        else:
                            self._sources.append(CatalogSourceInfo(filename=filename, build_datetime=build_datetime))

    def __bool__(self) -> bool:
        return bool(self._data.catalog)

    @property
    def is_empty(self) -> bool:
        return not bool(self)

    @property
    def sources(self) -> list[Path]:
        return [source.filename for source in self._sources]

    @property
    def sources_info(self) -> list[CatalogSourceInfo]:
        return self._sources.copy()

    @property
    def catalog(self) -> CatalogType:
        return self._data.catalog

    @property
    def entries_count(self) -> int:
        return len(self._data.catalog)

    @property
    def frequency_limits(self) -> tuple[tuple[float, float], ...]:
        return self._data.frequency_limits if self._data.catalog else (0.0, math.inf)

    @property
    def min_frequency(self) -> float:
        return min(min(f) for f in self._data.frequency_limits) if self._data.frequency_limits else 0.0

    @property
    def max_frequency(self) -> float:
        return max(max(f) for f in self._data.frequency_limits) if self._data.frequency_limits else math.inf

    def filter(
        self,
        *,
        min_frequency: float = 0.0,
        max_frequency: float = math.inf,
        min_intensity: float = -math.inf,
        max_intensity: float = math.inf,
        temperature: float = -math.inf,
        any_name: str = "",
        any_formula: str = "",
        any_name_or_formula: str = "",
        anything: str = "",
        species_tag: int = 0,
        inchi_key: str = "",
        trivial_name: str = "",
        structural_formula: str = "",
        name: str = "",
        stoichiometric_formula: str = "",
        isotopolog: str = "",
        state: str = "",
        degrees_of_freedom: int | None = None,
    ) -> CatalogType:
        """
        Extract only the entries that match all the specified conditions

        :param float min_frequency: The lower frequency [MHz] to take.
        :param float max_frequency: The upper frequency [MHz] to take.
        :param float min_intensity: The minimal intensity [log10(nm²×MHz)] to take.
        :param float max_intensity: The maximal intensity [log10(nm²×MHz)] to take, use to avoid meta-stable substances.
        :param float temperature: The temperature to calculate the line intensity at,
                                  use the catalog intensity if not set.
        :param str any_name: A string to match the “trivialname” or the “name” field.
        :param str any_formula: A string to match the “structuralformula,” “moleculesymbol,”
                                “stoichiometricformula,” or “isotopolog” field.
        :param str any_name_or_formula: A string to match any field used by :param:any_name and :param:any_formula.
        :param str anything: A string to match any field.
        :param int species_tag: A number to match the “speciestag” field.
        :param str inchi_key: A string to match the “inchikey” field.
                              See https://iupac.org/who-we-are/divisions/division-details/inchi/ for more.
        :param str trivial_name: A string to match the “trivialname” field.
        :param str structural_formula: A string to match the “structuralformula” field.
        :param str name: A string to match the “name” field.
        :param str stoichiometric_formula: A string to match the “stoichiometricformula” field.
        :param str isotopolog: A string to match the “isotopolog” field.
        :param str state: A string to match the “state” or the “state_html” field.
        :param int degrees_of_freedom: 0 for atoms, 2 for linear molecules, and 3 for nonlinear molecules.
        :return: A dict of substances with non-empty lists of absorption lines that match all the conditions.
        """

        if self.is_empty:
            return dict()

        if min_frequency > max_frequency or min_frequency > self.max_frequency or max_frequency < self.min_frequency:
            return dict()

        def check_str(pattern: str, *text: str) -> bool:
            return not pattern or any(pattern == t for t in text)

        st: int
        selected_entries: CatalogType = dict()
        entry: CatalogEntryType
        filtered_entry: CatalogEntryType
        if any(
            (
                species_tag,
                inchi_key,
                trivial_name,
                structural_formula,
                name,
                stoichiometric_formula,
                isotopolog,
                state,
                degrees_of_freedom,
                any_name,
                any_formula,
                any_name_or_formula,
                anything,
            )
        ):
            trivial_name: str = trivial_name.casefold()
            name: str = name.casefold()
            any_name: str = any_name.casefold()
            any_name_or_formula_lowercase: str = any_name_or_formula.casefold()
            anything_lowercase: str = anything.casefold()
            for st in self._data.catalog if not species_tag else [species_tag]:
                entry = self._data.catalog.get(st, dict())
                if not entry:
                    continue
                if all(
                    (
                        check_str(inchi_key, entry.inchikey),
                        check_str(trivial_name, entry.trivialname.casefold()),
                        check_str(structural_formula, entry.structuralformula),
                        check_str(name, entry.name.casefold()),
                        check_str(stoichiometric_formula, entry.stoichiometricformula),
                        check_str(isotopolog, entry.isotopolog),
                        check_str(state, entry.state, entry.state_html),
                        (degrees_of_freedom is None or entry.degreesoffreedom == degrees_of_freedom),
                        check_str(
                            any_name,
                            entry.trivialname.casefold(),
                            entry.name.casefold(),
                        ),
                        check_str(
                            any_formula,
                            entry.structuralformula,
                            entry.moleculesymbol,
                            entry.stoichiometricformula,
                            entry.isotopolog,
                        ),
                        (
                            not any_name_or_formula
                            or check_str(
                                any_name_or_formula_lowercase,
                                entry.trivialname.casefold(),
                                entry.name.casefold(),
                            )
                            or check_str(
                                any_name_or_formula,
                                entry.structuralformula,
                                entry.moleculesymbol,
                                entry.stoichiometricformula,
                                entry.isotopolog,
                            )
                        ),
                        (
                            not anything
                            or anything in (str(getattr(entry, key)) for key in entry.__slots__ if key != LINES)
                            or check_str(
                                anything_lowercase,
                                entry.trivialname.casefold(),
                                entry.name.casefold(),
                            )
                        ),
                    )
                ):
                    filtered_entry = filter_by_frequency_and_intensity(
                        entry,
                        temperature=temperature,
                        min_frequency=min_frequency,
                        max_frequency=max_frequency,
                        min_intensity=min_intensity,
                        max_intensity=max_intensity,
                    )
                    if filtered_entry.lines:
                        selected_entries[st] = filtered_entry
        else:
            for st in self._data.catalog:
                entry = self.catalog.get(st, dict())
                if not entry:
                    continue
                filtered_entry = filter_by_frequency_and_intensity(
                    entry,
                    temperature=temperature,
                    min_frequency=min_frequency,
                    max_frequency=max_frequency,
                    min_intensity=min_intensity,
                    max_intensity=max_intensity,
                )
                if filtered_entry.lines:
                    selected_entries[st] = filtered_entry
        return selected_entries

    def filter_by_species_tags(
        self,
        *,
        species_tags: Iterable[int] | None = None,
        min_frequency: float = 0.0,
        max_frequency: float = math.inf,
        min_intensity: float = -math.inf,
        max_intensity: float = math.inf,
        temperature: float = -math.inf,
    ) -> CatalogType:
        """
        Extract only the entries that match the specified conditions

        :param Iterable[int] species_tags: Numbers to match the “speciestag” field.
        :param float min_frequency: The lower frequency [MHz] to take.
        :param float max_frequency: The upper frequency [MHz] to take.
        :param float min_intensity: The minimal intensity [log10(nm²×MHz)] to take.
        :param float max_intensity: The maximal intensity [log10(nm²×MHz)] to take, use to avoid meta-stable substances.
        :param float temperature: The temperature to calculate the line intensity at,
                                  use the catalog intensity if not set.
        :return: A dict of substances with non-empty lists of absorption lines that match all the conditions.
        """

        if self.is_empty:
            return dict()

        if min_frequency > max_frequency or min_frequency > self.max_frequency or max_frequency < self.min_frequency:
            return dict()

        species_tag: int
        selected_entries: CatalogType = dict()
        entry: CatalogEntryType
        filtered_entry: CatalogEntryType
        for species_tag in species_tags if species_tags is not None else self._data.catalog:
            entry = self.catalog.get(species_tag, dict())
            if not entry:
                continue
            filtered_entry = filter_by_frequency_and_intensity(
                entry,
                temperature=temperature,
                min_frequency=min_frequency,
                max_frequency=max_frequency,
                min_intensity=min_intensity,
                max_intensity=max_intensity,
            )
            if filtered_entry.lines:
                selected_entries[species_tag] = filtered_entry
        return selected_entries

    def print(self, **kwargs: None | int | float | str) -> None:
        """
        Print a table of the filtered catalog entries

        :param kwargs: All arguments that are valid for :func:`filter <catalog.Catalog.filter>`
        :return: nothing
        """
        entries: CatalogType = self.filter(**kwargs)
        if not entries:
            print("nothing found")
            return

        names: list[str] = []
        frequencies: list[float] = []
        intensities: list[float] = []
        entry: CatalogEntryType
        for species_tag in entries:
            entry = entries[species_tag]
            for line in entry.lines:
                names.append(entry.name)
                frequencies.append(line.frequency)
                intensities.append(line.intensity)

        def max_width(items: list[str]) -> int:
            return max(len(item) for item in items)

        def max_precision(items: list[str]) -> int:
            return max((len(item) - item.find(".")) for item in items) - 1

        names_width: int = max_width(names)
        frequencies_str: list[str] = list(map(str, frequencies))
        intensities_str: list[str] = list(map(str, intensities))
        frequencies_width: int = max_width(frequencies_str)
        intensities_width: int = max_width(intensities_str)
        frequencies_precision: int = max_precision(frequencies_str)
        intensities_precision: int = max_precision(intensities_str)
        for j, (n, f, i) in enumerate(zip(names, frequencies, intensities)):
            print(
                " ".join(
                    (
                        f"{n:<{names_width}}",
                        f"{f:>{frequencies_width}.{frequencies_precision}f}",
                        f"{i:>{intensities_width}.{intensities_precision}f}",
                    )
                )
            )

    @classmethod
    def from_data(
        cls,
        catalog_data: CatalogType | CatalogJSONType,
        frequency_limits: tuple[float, float] = (0.0, math.inf),
    ) -> "Catalog":
        catalog: Catalog = Catalog()
        catalog._data.catalog = catalog_data
        catalog._data.frequency_limits = (frequency_limits,)
        return catalog

    def save(self, filename: str | PathLike[str], build_time: datetime = datetime.now(tz=timezone.utc)) -> None:
        opener: Catalog.Opener
        try:
            opener = Catalog.Opener(filename)
        except ValueError:
            filename = Path(filename)
            opener = Catalog.Opener(filename.with_name(filename.name + Catalog.DEFAULT_SUFFIX))

        def _repr(o: object) -> str:
            if isinstance(o, str):
                return '"' + o.replace('"', r"\"") + '"'
            if isinstance(o, Iterable):
                return "[" + ",".join(_repr(sub_o) for sub_o in o) + "]"
            if isinstance(o, Mapping):
                return "{" + ",".join(":".join((_repr(key), _repr(value))) for key, value in o.items()) + "}"
            if isinstance(o, (CatalogEntryType, LineType)):
                return "{" + ",".join(":".join((_repr(key), _repr(getattr(o, key)))) for key in o.__slots__) + "}"
            return repr(o)

        f: TextIO
        with opener.open("wt") as f:
            f.write("{")
            f.write(_repr(CATALOG))
            f.write(":{")
            is_not_first_item: bool = False
            for species_tag in self._data.catalog:
                if is_not_first_item:
                    f.write(",")
                else:
                    is_not_first_item = True
                f.write(_repr(str(species_tag)))
                f.write(":")
                f.write(_repr(self._data.catalog[species_tag]))
            f.write("},")
            f.write(_repr(FREQUENCY))
            f.write(":")
            f.write(_repr(self._data.frequency_limits))
            f.write(",")
            f.write(_repr(BUILD_TIME))
            f.write(":")
            f.write(_repr(build_time.isoformat()))
            f.write("}")
